from typing import Any, AnyStr


class DltException(Exception):
    pass


class SignalReceivedException(DltException):
    def __init__(self, signal_code: int) -> None:
        self.signal_code = signal_code
        super().__init__(f"Signal {signal_code} received")


class PoolException(DltException):
    """
    Thrown by worker pool to pass information when thrown during processing an item
    """
    def __init__(self, pool_name: str = None, item: str = None, internal_exception: Exception = None) -> None:
        # we need it to make it pickle compatible
        if pool_name:
            self.pool_name = pool_name
            self.item = item
            self.internal_exception = internal_exception
            super().__init__(f"Pool {pool_name} raised on item {item} with {str(internal_exception)}")


class UnsupportedProcessStartMethodException(DltException):
    def __init__(self, method: str) -> None:
        self.method = method
        super().__init__(f"Process pool supports only fork start method, {method} not supported. Switch the pool type to threading")


class CannotInstallDependency(DltException):
    def __init__(self, dependency: str, interpreter: str, output: AnyStr) -> None:
        self.dependency = dependency
        self.interpreter = interpreter
        if isinstance(output, bytes):
            str_output = output.decode("utf-8")
        else:
            str_output = output
        super().__init__(f"Cannot install dependency {dependency} with {interpreter} and pip:\n{str_output}\n")


class VenvNotFound(DltException):
    def __init__(self, interpreter: str) -> None:
        self.interpreter = interpreter
        super().__init__(f"Venv with interpreter {interpreter} not found in path")


class TerminalException(Exception):
    """
    Marks an exception that cannot be recovered from, should be mixed in into concrete exception class
    """
    pass


class TransientException(Exception):
    """
    Marks an exception in operation that can be retried, should be mixed in into concrete exception class
    """
    pass


class TerminalValueError(ValueError, TerminalException):
    """
    ValueError that is unrecoverable
    """
    pass


class TimeRangeExhaustedException(DltException):
    """
    Raised when backfilling complete and no more time ranges can be generated
    """
    def __init__(self, start_ts: float, end_ts: float) -> None:
        self.start_ts = start_ts
        self.end_ts = end_ts
        super().__init__(f"Timerange ({start_ts} to {end_ts}> exhausted")


class DictValidationException(DltException):
    def __init__(self, msg: str, path: str, field: str = None, value: Any = None):
        self.path = path
        self.field = field
        self.value = value
        super().__init__(msg)



class SchemaVolumeConfiguration:
    SCHEMA_VOLUME_PATH: str = "_storage/schemas"  # path to volume with default schemas


class ProductionSchemaVolumeConfiguration:
    SCHEMA_VOLUME_PATH: str = None



import random
from types import TracebackType
from typing import Dict, Literal, Type
from dlt.common.dataset_writers import TWriterType

from dlt.common import pendulum
from dlt.common.schema import Schema

from dlt.loaders.client_base import JobClientBase, LoadJob, TJobClientCapabilities
from dlt.loaders.local_types import LoadJobStatus
from dlt.loaders.exceptions import (LoadJobNotExistsException, LoadJobInvalidStateTransitionException,
                                            LoadClientTerminalException, LoadClientTransientException)

from dlt.loaders.dummy.configuration import DummyClientConfiguration


class LoadDummyJob(LoadJob):
    def __init__(self, file_name: str, fail_prob: float = 0.0, retry_prob: float = 0.0, completed_prob: float = 1.0, timeout: float = 10.0) -> None:
        self.fail_prob = fail_prob
        self.retry_prob = retry_prob
        self.completed_prob = completed_prob
        self.timeout = timeout
        self._status: LoadJobStatus = "running"
        self._exception: str = None
        self.start_time: float = pendulum.now().timestamp()
        super().__init__(file_name)
        s = self.status()
        if s == "failed":
            raise LoadClientTerminalException(self._exception)
        if s == "retry":
            raise LoadClientTransientException(self._exception)


    def status(self) -> LoadJobStatus:
        # this should poll the server for a job status, here we simulate various outcomes
        if self._status == "running":
            n = pendulum.now().timestamp()
            if n - self.start_time > self.timeout:
                self._status = "failed"
                self._exception = "failed due to timeout"
            else:
                c_r = random.random()
                if self.completed_prob >= c_r:
                    self._status = "completed"
                else:
                    c_r = random.random()
                    if self.retry_prob >= c_r:
                        self._status = "retry"
                        self._exception = "a random retry occured"
                    else:
                        c_r = random.random()
                        if self.fail_prob >= c_r:
                            self._status = "failed"
                            self._exception = "a random fail occured"

        return self._status

    def file_name(self) -> str:
        return self._file_name

    def exception(self) -> str:
        # this will typically call server for error messages
        return self._exception

    def retry(self) -> None:
        if self._status != "retry":
            raise LoadJobInvalidStateTransitionException(self._status, "retry")
        self._status = "retry"


JOBS: Dict[str, LoadDummyJob] = {}


class DummyClient(JobClientBase):
    """
    dummy client storing jobs in memory
    """
    def __init__(self, schema: Schema, CONFIG: Type[DummyClientConfiguration]) -> None:
        self.C = CONFIG
        super().__init__(schema)

    def initialize_storage(self) -> None:
        pass

    def update_storage_schema(self) -> None:
        pass

    def start_file_load(self, table_name: str, file_path: str) -> LoadJob:
        self._get_table_by_name(table_name, file_path)
        job_id = JobClientBase.get_file_name_from_file_path(file_path)
        file_name = JobClientBase.get_file_name_from_file_path(file_path)
        # return existing job if already there
        if job_id not in JOBS:
            JOBS[job_id] = self._create_job(file_name)
        else:
            job = JOBS[job_id]
            if job.status == "retry":
                job.retry()

        return JOBS[job_id]

    def restore_file_load(self, file_path: str) -> LoadJob:
        job_id = JobClientBase.get_file_name_from_file_path(file_path)
        if job_id not in JOBS:
            raise LoadJobNotExistsException(job_id)
        return JOBS[job_id]

    def complete_load(self, load_id: str) -> None:
        pass

    @property
    def capabilities(self) -> TJobClientCapabilities:
        return {
            "writer_type": supported_writer(self.C)
        }

    def __enter__(self) -> "DummyClient":
        return self

    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType) -> None:
        pass

    def _create_job(self, job_id: str) -> LoadDummyJob:
        return LoadDummyJob(
            job_id,
            fail_prob=self.C.FAIL_PROB,
            retry_prob=self.C.RETRY_PROB,
            completed_prob=self.C.COMPLETED_PROB,
            timeout=self.C.TIMEOUT
            )



def make_client(schema: Schema, C: Type[DummyClientConfiguration]) -> JobClientBase:
    return DummyClient(schema, C)


def supported_writer(C: Type[DummyClientConfiguration]) -> TWriterType:
    return C.WRITER_TYPE



from dlt.common.typing import StrStr
from dlt.common.configuration.utils import TConfigSecret

class GcpClientConfiguration:
    PROJECT_ID: str = None
    DATASET: str = None
    TIMEOUT: float = 30.0
    BQ_CRED_TYPE: str = "service_account"
    BQ_CRED_PRIVATE_KEY: TConfigSecret = None
    BQ_CRED_TOKEN_URI: str = "https://oauth2.googleapis.com/token"
    BQ_CRED_CLIENT_EMAIL: str = None

    @classmethod
    def check_integrity(cls) -> None:
        if cls.BQ_CRED_PRIVATE_KEY and cls.BQ_CRED_PRIVATE_KEY[-1] != "\n":
            # must end with new line, otherwise won't be parsed by Crypto
            cls.BQ_CRED_PRIVATE_KEY = TConfigSecret(cls.BQ_CRED_PRIVATE_KEY + "\n")

    @classmethod
    def to_service_credentials(cls) -> StrStr:
        return {
                "type": cls.BQ_CRED_TYPE,
                "project_id": cls.PROJECT_ID,
                "private_key": cls.BQ_CRED_PRIVATE_KEY,
                "token_uri": cls.BQ_CRED_TOKEN_URI,
                "client_email": cls.BQ_CRED_CLIENT_EMAIL
            }


class GcpClientProductionConfiguration(GcpClientConfiguration):
    PROJECT_ID: str = None
    DATASET: str = None
    BQ_CRED_PRIVATE_KEY: TConfigSecret = None
    BQ_CRED_CLIENT_EMAIL: str = None



from copy import deepcopy
import re
import base64
import binascii
import datetime  # noqa: I251
from typing import Dict, List, Sequence, Type, Any, cast

from dlt.common import pendulum, json, Decimal
from dlt.common.arithmetics import ConversionSyntax
from dlt.common.exceptions import DictValidationException
from dlt.common.normalizers.names import TNormalizeNameFunc
from dlt.common.typing import DictStrAny, REPattern
from dlt.common.validation import TCustomValidator, validate_dict
from dlt.common.schema import detections
from dlt.common.schema.typing import SIMPLE_REGEX_PREFIX, TColumnName, TNormalizersConfig, TSimpleRegex, TStoredSchema, TTable, TTableColumns, TColumnBase, TColumn, TColumnProp, TDataType, THintType, TTypeDetectionFunc, TTypeDetections
from dlt.common.schema.exceptions import ParentTableNotFoundException, SchemaEngineNoUpgradePathException


RE_LEADING_DIGITS = re.compile(r"^\d+")
RE_NON_ALPHANUMERIC = re.compile(r"[^a-zA-Z\d]")
RE_NON_ALPHANUMERIC_UNDERSCORE = re.compile(r"[^a-zA-Z\d_]")


# fix a name so it is acceptable as schema name
def normalize_schema_name(name: str) -> str:
    if name is None:
        raise ValueError(name)

    # prefix the name starting with digits
    if RE_LEADING_DIGITS.match(name):
        name = "s" + name
    # leave only alphanumeric
    return RE_NON_ALPHANUMERIC.sub("", name).lower()


def apply_defaults(stored_schema: TStoredSchema) -> None:
    for table_name, table in stored_schema["tables"].items():
        # overwrite name
        table["name"] = table_name
        # add default write disposition to root tables
        if table.get("parent") is None and table.get("write_disposition") is None:
            table["write_disposition"] = "append"
        # add missing hints to columns
        for column_name in table["columns"]:
            # add default hints to tables
            column = add_missing_hints(table["columns"][column_name])
            # overwrite column name
            column["name"] = column_name
            # set column with default
            table["columns"][column_name] = column


def remove_defaults(stored_schema: TStoredSchema) -> None:
    clean_tables = deepcopy(stored_schema["tables"])
    for t in clean_tables.values():
        del t["name"]
        for c in t["columns"].values():
            # do not save names
            del c["name"]  # type: ignore
            # remove hints with default values
            for h in list(c.keys()):
                if isinstance(c[h], bool) and c[h] is False and h != "nullable":  # type: ignore
                    del c[h]  # type: ignore

    stored_schema["tables"] = clean_tables


def simple_regex_validator(path: str, pk: str, pv: Any, t: Any) -> bool:
    # custom validator on type TSimpleRegex
    if t is TSimpleRegex:
        if not isinstance(pv, str):
            raise DictValidationException(f"In {path}: field {pk} value {pv} has invalid type {type(pv).__name__} while str is expected", path, pk, pv)
        if pv.startswith(SIMPLE_REGEX_PREFIX):
            # check if regex
            try:
                re.compile(pv[3:])
            except Exception as e:
                raise DictValidationException(f"In {path}: field {pk} value {pv[3:]} does not compile as regex: {str(e)}", path, pk, pv)
        else:
            if RE_NON_ALPHANUMERIC_UNDERSCORE.match(pv):
                raise DictValidationException(f"In {path}: field {pk} value {pv} looks like a regex, please prefix with re:", path, pk, pv)
        # we know how to validate that type
        return True
    else:
        # don't know how to validate t
        return False


def column_name_validator(normalize_func: TNormalizeNameFunc) -> TCustomValidator:

    def validator(path: str, pk: str, pv: Any, t: Any) -> bool:
        if t is TColumnName:
            if not isinstance(pv, str):
                raise DictValidationException(f"In {path}: field {pk} value {pv} has invalid type {type(pv).__name__} while str is expected", path, pk, pv)
            if normalize_func(pv) != pv:
                raise DictValidationException(f"In {path}: field {pk}: {pv} is not a valid column name", path, pk, pv)
            return True
        else:
            return False

    return validator


def compile_simple_regex(r: TSimpleRegex) -> REPattern:
    if r.startswith(SIMPLE_REGEX_PREFIX):
        return re.compile(r[3:])
    else:
        # exact matches
        return re.compile("^" + re.escape(r) + "$")


def validate_stored_schema(stored_schema: TStoredSchema) -> None:
    # use lambda to verify only non extra fields
    validate_dict(TStoredSchema, stored_schema, ".", lambda k: not k.startswith("x-"), simple_regex_validator)
    # check child parent relationships
    for table_name, table in stored_schema["tables"].items():
        parent_table_name = table.get("parent")
        if parent_table_name:
            if parent_table_name not in stored_schema["tables"]:
                raise ParentTableNotFoundException(table_name, parent_table_name)


def upgrade_engine_version(schema_dict: DictStrAny, from_engine: int, to_engine: int) -> TStoredSchema:
    if from_engine == to_engine:
        return cast(TStoredSchema, schema_dict)

    if from_engine == 1 and to_engine > 1:
        schema_dict["engine_version"] = 2
        schema_dict["includes"] = []
        schema_dict["excludes"] = []
        from_engine = 2
    if from_engine == 2 and to_engine > 2:
        # current version of the schema
        current = cast(TStoredSchema, schema_dict)
        # add default normalizers and root hash propagation
        current["normalizers"] = default_normalizers()
        current["normalizers"]["json"]["config"] = {
                    "propagation": {
                        "root": {
                            "_dlt_id": "_dlt_root_id"
                        }
                    }
                }
        # move settings, convert strings to simple regexes
        d_h: Dict[THintType, List[TSimpleRegex]] = schema_dict.pop("hints", {})
        for h_k, h_l in d_h.items():
            d_h[h_k] = list(map(lambda r: TSimpleRegex("re:" + r), h_l))
        p_t: Dict[TSimpleRegex, TDataType] = schema_dict.pop("preferred_types", {})
        p_t = {TSimpleRegex("re:" + k): v for k, v in p_t.items()}

        current["settings"] = {
            "default_hints": d_h,
            "preferred_types": p_t,
        }
        # repackage tables
        old_tables: Dict[str, TTableColumns] = schema_dict.pop("tables")
        current["tables"] = {}
        for name, columns in old_tables.items():
            # find last path separator
            parent = name
            # go back in a loop to find existing parent
            while True:
                idx = parent.rfind("__")
                if idx > 0:
                    parent = parent[:idx]
                    if parent not in old_tables:
                        # print(f"candidate {parent} for {name} not found")
                        continue
                else:
                    parent = None
                break
            nt = new_table(name, parent)
            nt["columns"] = columns
            current["tables"][name] = nt
        # assign exclude and include to tables

        def migrate_filters(group: str, filters: List[str]) -> None:
            # existing filter were always defined at the root table. find this table and move filters
            for f in filters:
                # skip initial ^
                root = f[1:f.find("__")]
                path = f[f.find("__") + 2:]
                t = current["tables"].get(root)
                if t is None:
                    # must add new table to hold filters
                    t = new_table(root)
                    current["tables"][root] = t
                t.setdefault("filters", {}).setdefault(group, []).append("re:^" + path)  # type: ignore

        excludes = schema_dict.pop("excludes", [])
        migrate_filters("excludes", excludes)
        includes = schema_dict.pop("includes", [])
        migrate_filters("includes", includes)

        # upgraded
        schema_dict["engine_version"] = 3
        from_engine = 3
    if from_engine == 3:
        pass
    if from_engine != to_engine:
        raise SchemaEngineNoUpgradePathException(schema_dict["name"], schema_dict["engine_version"], from_engine, to_engine)
    return cast(TStoredSchema, schema_dict)


def add_missing_hints(column: TColumnBase) -> TColumn:
    return {
        **{  # type:ignore
            "partition": False,
            "cluster": False,
            "unique": False,
            "sort": False,
            "primary_key": False,
            "foreign_key": False,
        },
        **column
    }


def autodetect_sc_type(detection_fs: Sequence[TTypeDetections], t: Type[Any], v: Any) -> TDataType:
    if detection_fs:
        for detection_fn in detection_fs:
            # the method must exist in the module
            detection_f: TTypeDetectionFunc = getattr(detections, "is_" + detection_fn)
            dt = detection_f(t, v)
            if dt is not None:
                return dt
    return None


def py_type_to_sc_type(t: Type[Any]) -> TDataType:
    if t is float:
        return "double"
    elif t is int:
        return "bigint"
    elif t is bool:
        return "bool"
    elif issubclass(t, bytes):
        return "binary"
    elif issubclass(t, dict) or issubclass(t, list):
        return "complex"
    elif issubclass(t, Decimal):
        return "decimal"
    elif issubclass(t, datetime.datetime):
        return "timestamp"
    else:
        return "text"


def coerce_type(to_type: TDataType, from_type: TDataType, value: Any) -> Any:
    if to_type == from_type:
        if to_type == "complex":
            # complex types will be always represented as strings
            return json.dumps(value)
        return value

    if to_type == "text":
        if from_type == "complex":
            return json.dumps(value)
        else:
            return str(value)

    if to_type == "binary":
        if from_type == "text":
            if value.startswith("0x"):
                return bytes.fromhex(value[2:])
            try:
                return base64.b64decode(value, validate=True)
            except binascii.Error:
                raise ValueError(value)
        if from_type == "bigint":
            return value.to_bytes((value.bit_length() + 7) // 8, 'little')

    if to_type in ["wei", "bigint"]:
        if from_type == "bigint":
            return value
        if from_type in ["decimal", "double"]:
            if value % 1 != 0:
                # only integer decimals and floats can be coerced
                raise ValueError(value)
            return int(value)
        if from_type == "text":
            trim_value = value.strip()
            if trim_value.startswith("0x"):
                return int(trim_value[2:], 16)
            else:
                return int(trim_value)

    if to_type == "double":
        if from_type in ["bigint", "wei", "decimal"]:
            return float(value)
        if from_type == "text":
            trim_value = value.strip()
            if trim_value.startswith("0x"):
                return float(int(trim_value[2:], 16))
            else:
                return float(trim_value)

    if to_type == "decimal":
        if from_type in ["bigint", "wei"]:
            return value
        if from_type == "double":
            return Decimal(value)
        if from_type == "text":
            trim_value = value.strip()
            if trim_value.startswith("0x"):
                return int(trim_value[2:], 16)
            elif "." not in trim_value and "e" not in trim_value:
                return int(trim_value)
            else:
                try:
                    return Decimal(trim_value)
                except ConversionSyntax:
                    raise ValueError(trim_value)

    if to_type == "timestamp":
        if from_type in ["bigint", "double"]:
            # returns ISO datetime with timezone
            return pendulum.from_timestamp(value)

        if from_type == "text":
            # if parses as ISO date then pass it
            try:
                dtv = pendulum.parse(value, strict=False, exact=True)
                if isinstance(dtv, datetime.time):
                    raise ValueError(value)
                return dtv
            except ValueError:
                # try to convert string to integer, or float
                try:
                    value = int(value)
                except ValueError:
                    # raises ValueError if not parsing correctly
                    value = float(value)
                return pendulum.from_timestamp(value)

    raise ValueError(value)


def compare_columns(a: TColumn, b: TColumn) -> bool:
    return a["data_type"] == b["data_type"] and a["nullable"] == b["nullable"]


def hint_to_column_prop(h: THintType) -> TColumnProp:
    if h == "not_null":
        return "nullable"
    return h


def version_table() -> TTable:
    table = new_table("_dlt_version", columns=[
            add_missing_hints({
                "name": "version",
                "data_type": "bigint",
                "nullable": False,
            }),
            add_missing_hints({
                "name": "engine_version",
                "data_type": "bigint",
                "nullable": False
            }),
            add_missing_hints({
                "name": "inserted_at",
                "data_type": "timestamp",
                "nullable": False
            })
        ]
    )
    table["write_disposition"] = "skip"
    table["description"] = "Created by DLT. Tracks schema updates"
    return table


def load_table() -> TTable:
    table = new_table("_dlt_loads", columns=[
            add_missing_hints({
                "name": "load_id",
                "data_type": "text",
                "nullable": False
            }),
            add_missing_hints({
                "name": "status",
                "data_type": "bigint",
                "nullable": False
            }),
            add_missing_hints({
                "name": "inserted_at",
                "data_type": "timestamp",
                "nullable": False
            })
        ]
    )
    table["write_disposition"] = "skip"
    table["description"] = "Created by DLT. Tracks completed loads"
    return table


def new_table(table_name: str, parent_name: str = None, columns: Sequence[TColumn] = None) -> TTable:
    table: TTable = {
        "name": table_name,
        "columns": {} if columns is None else {c["name"]: c for c in columns}
    }
    if parent_name:
        table["parent"] = parent_name
    else:
        # set write disposition only for root tables
        table["write_disposition"] = "append"
    return table


def default_normalizers() -> TNormalizersConfig:
    return {
                "detections": ["timestamp", "iso_timestamp"],
                "names": "dlt.common.normalizers.names.snake_case",
                "json": {
                    "module": "dlt.common.normalizers.json.relational"
                }
            }


def standard_hints() -> Dict[THintType, List[TSimpleRegex]]:
    return None



import os
from pathlib import Path
from typing import List, Literal, Optional, Sequence, Tuple, Type

from dlt.common import json, pendulum
from dlt.common.file_storage import FileStorage
from dlt.common.dataset_writers import TWriterType, write_jsonl, write_insert_values
from dlt.common.configuration import LoadingVolumeConfiguration
from dlt.common.exceptions import TerminalValueError
from dlt.common.schema import TSchemaUpdate, TTableColumns
from dlt.common.storages.versioned_storage import VersionedStorage
from dlt.common.typing import StrAny

from dlt.common.storages.exceptions import StorageException


# folders to manage load jobs in a single load package
TWorkingFolder = Literal["new_jobs", "failed_jobs", "started_jobs", "completed_jobs"]

class LoaderStorage(VersionedStorage):

    STORAGE_VERSION = "1.0.0"
    LOADING_FOLDER = "loading"  # folder within the volume where load packages are stored
    LOADED_FOLDER = "loaded"  # folder to keep the loads that were completely processed

    NEW_JOBS_FOLDER: TWorkingFolder = "new_jobs"
    FAILED_JOBS_FOLDER: TWorkingFolder = "failed_jobs"
    STARTED_JOBS_FOLDER: TWorkingFolder = "started_jobs"
    COMPLETED_JOBS_FOLDER: TWorkingFolder = "completed_jobs"

    LOAD_SCHEMA_UPDATE_FILE_NAME = "schema_updates.json"

    SUPPORTED_WRITERS: List[TWriterType] = ["jsonl", "insert_values"]

    def __init__(self, is_owner: bool,  C: Type[LoadingVolumeConfiguration], writer_type: TWriterType) -> None:
        if writer_type not in LoaderStorage.SUPPORTED_WRITERS:
            raise TerminalValueError(writer_type)
        self.writer_type = writer_type
        self.delete_completed_jobs = C.DELETE_COMPLETED_JOBS
        super().__init__(LoaderStorage.STORAGE_VERSION, is_owner, FileStorage(C.LOADING_VOLUME_PATH, "t", makedirs=is_owner))

    def initialize_storage(self) -> None:
        self.storage.create_folder(LoaderStorage.LOADED_FOLDER, exists_ok=True)
        self.storage.create_folder(LoaderStorage.LOADING_FOLDER, exists_ok=True)

    def create_temp_load_folder(self, load_id: str) -> None:
        # delete previous version
        if self.storage.has_folder(load_id):
            self.storage.delete_folder(load_id, recursively=True)
        self.storage.create_folder(load_id)
        # create processing directories
        self.storage.create_folder(f"{load_id}/{LoaderStorage.NEW_JOBS_FOLDER}")
        self.storage.create_folder(f"{load_id}/{LoaderStorage.COMPLETED_JOBS_FOLDER}")
        self.storage.create_folder(f"{load_id}/{LoaderStorage.FAILED_JOBS_FOLDER}")
        self.storage.create_folder(f"{load_id}/{LoaderStorage.STARTED_JOBS_FOLDER}")

    def write_temp_loading_file(self, load_id: str, table_name: str, table: TTableColumns, file_id: str, rows: Sequence[StrAny]) -> str:
        file_name = self.build_loading_file_name(load_id, table_name, file_id)
        with self.storage.open_file(file_name, mode="w") as f:
            if self.writer_type == "jsonl":
                write_jsonl(f, rows)
            elif self.writer_type == "insert_values":
                write_insert_values(f, rows, table.keys())
        return Path(file_name).name

    def save_schema_updates(self, load_id: str, schema_updates: Sequence[TSchemaUpdate]) -> None:
        with self.storage.open_file(f"{load_id}/{LoaderStorage.LOAD_SCHEMA_UPDATE_FILE_NAME}", mode="w") as f:
            json.dump(schema_updates, f)

    def commit_temp_load_folder(self, load_id: str) -> None:
        self.storage.atomic_rename(load_id, self.get_load_path(load_id))

    def list_loads(self) -> Sequence[str]:
        loads = self.storage.list_folder_dirs(LoaderStorage.LOADING_FOLDER, to_root=False)
        # start from the oldest packages
        return sorted(loads)

    def list_completed_loads(self) -> Sequence[str]:
        loads = self.storage.list_folder_dirs(LoaderStorage.LOADED_FOLDER, to_root=False)
        # start from the oldest packages
        return sorted(loads)

    def list_new_jobs(self, load_id: str) -> Sequence[str]:
        new_jobs = self.storage.list_folder_files(f"{self.get_load_path(load_id)}/{LoaderStorage.NEW_JOBS_FOLDER}")
        # make sure all jobs have supported writers
        wrong_job = next((j for j in new_jobs if LoaderStorage.parse_load_file_name(j)[1] != self.writer_type), None)
        if wrong_job is not None:
            raise JobWithUnsupportedWriterException(load_id, self.writer_type, wrong_job)
        return new_jobs

    def list_started_jobs(self, load_id: str) -> Sequence[str]:
        return self.storage.list_folder_files(f"{self.get_load_path(load_id)}/{LoaderStorage.STARTED_JOBS_FOLDER}")

    def list_failed_jobs(self, load_id: str) -> Sequence[str]:
        return self.storage.list_folder_files(f"{self.get_load_path(load_id)}/{LoaderStorage.FAILED_JOBS_FOLDER}")

    def list_archived_failed_jobs(self, load_id: str) -> Sequence[str]:
        return self.storage.list_folder_files(f"{self.get_archived_path(load_id)}/{LoaderStorage.FAILED_JOBS_FOLDER}")

    def begin_schema_update(self, load_id: str) -> Optional[TSchemaUpdate]:
        schema_update_file = f"{self.get_load_path(load_id)}/{LoaderStorage.LOAD_SCHEMA_UPDATE_FILE_NAME}"
        if self.storage.has_file(schema_update_file):
            schema_update: TSchemaUpdate = json.loads(self.storage.load(schema_update_file))
            return schema_update
        else:
            return None

    def commit_schema_update(self, load_id: str) -> None:
        load_path = self.get_load_path(load_id)
        schema_update_file = f"{load_path}/{LoaderStorage.LOAD_SCHEMA_UPDATE_FILE_NAME}"
        self.storage.atomic_rename(schema_update_file, f"{load_path}/{LoaderStorage.COMPLETED_JOBS_FOLDER}/{LoaderStorage.LOAD_SCHEMA_UPDATE_FILE_NAME}")

    def start_job(self, load_id: str, file_name: str) -> str:
        return self._move_file(load_id, LoaderStorage.NEW_JOBS_FOLDER, LoaderStorage.STARTED_JOBS_FOLDER, file_name)

    def fail_job(self, load_id: str, file_name: str, failed_message: Optional[str]) -> str:
        load_path = self.get_load_path(load_id)
        if failed_message:
            self.storage.save(f"{load_path}/{LoaderStorage.FAILED_JOBS_FOLDER}/{file_name}.exception", failed_message)
        # move to failed jobs
        return self._move_file(load_id, LoaderStorage.STARTED_JOBS_FOLDER, LoaderStorage.FAILED_JOBS_FOLDER, file_name)

    def retry_job(self, load_id: str, file_name: str) -> str:
        return self._move_file(load_id, LoaderStorage.STARTED_JOBS_FOLDER, LoaderStorage.NEW_JOBS_FOLDER, file_name)

    def complete_job(self, load_id: str, file_name: str) -> str:
        return self._move_file(load_id, LoaderStorage.STARTED_JOBS_FOLDER, LoaderStorage.COMPLETED_JOBS_FOLDER, file_name)

    def archive_load(self, load_id: str) -> None:
        load_path = self.get_load_path(load_id)
        has_failed_jobs = len(self.list_failed_jobs(load_id)) > 0
        # delete load that does not contain failed jobs
        if self.delete_completed_jobs and not has_failed_jobs:
            self.storage.delete_folder(load_path, recursively=True)
        else:
            archive_path = self.get_archived_path(load_id)
            self.storage.atomic_rename(load_path, archive_path)

    def get_load_path(self, load_id: str) -> str:
        return f"{LoaderStorage.LOADING_FOLDER}/{load_id}"

    def get_archived_path(self, load_id: str) -> str:
        return f"{LoaderStorage.LOADED_FOLDER}/{load_id}"

    def build_loading_file_name(self, load_id: str, table_name: str, file_id: str) -> str:
        file_name = f"{table_name}.{file_id}.{self.writer_type}"
        return f"{load_id}/{LoaderStorage.NEW_JOBS_FOLDER}/{file_name}"

    def _move_file(self, load_id: str, source_folder: TWorkingFolder, dest_folder: TWorkingFolder, file_name: str) -> str:
        load_path = self.get_load_path(load_id)
        dest_path = f"{load_path}/{dest_folder}/{file_name}"
        self.storage.atomic_rename(f"{load_path}/{source_folder}/{file_name}", dest_path)
        return self.storage._make_path(dest_path)

    def job_elapsed_time_seconds(self, file_path: str) -> float:
        return pendulum.now().timestamp() - os.path.getmtime(file_path)  # type: ignore

    def _get_file_path(self, load_id: str, folder: TWorkingFolder, file_name: str) -> str:
        load_path = self.get_load_path(load_id)
        return f"{load_path}/{folder}/{file_name}"

    @staticmethod
    def parse_load_file_name(file_name: str) -> Tuple[str, TWriterType]:
        p = Path(file_name)
        ext: TWriterType = p.suffix[1:]  # type: ignore
        if ext not in LoaderStorage.SUPPORTED_WRITERS:
            raise TerminalValueError(ext)

        parts = p.stem.split(".")
        return (parts[0], ext)


class LoaderStorageException(StorageException):
    pass


class JobWithUnsupportedWriterException(LoaderStorageException):
    def __init__(self, load_id: str, expected_writer_type: TWriterType, wrong_job: str) -> None:
        self.load_id = load_id
        self.expected_writer_type = expected_writer_type
        self.wrong_job = wrong_job



import os
import base64
from contextlib import contextmanager
import hashlib
from os import environ
import secrets
from typing import Any, Iterator, Optional, Sequence, TypeVar, Mapping, List, TypedDict, Union

from dlt.common.typing import StrAny, DictStrAny, StrStr

T = TypeVar("T")


def chunks(seq: Sequence[T], n: int) -> Iterator[Sequence[T]]:
    for i in range(0, len(seq), n):
        yield seq[i:i + n]


def uniq_id() -> str:
    return secrets.token_hex(16)


def digest128(v: str) -> str:
    return base64.b64encode(hashlib.shake_128(v.encode("utf-8")).digest(15)).decode('ascii')


def str2bool(v: str) -> bool:
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise ValueError('Boolean value expected.')


def flatten_list_of_dicts(dicts: Sequence[StrAny]) -> StrAny:
    """
    Transforms a list of objects [{K: {...}}, {L: {....}}, ...] -> {K: {...}, L: {...}...}
    """
    o: DictStrAny = {}
    for d in dicts:
        for k,v in d.items():
            if k in o:
                raise KeyError(f"Cannot flatten with duplicate key {k}")
            o[k] = v
    return o


def flatten_list_of_str_or_dicts(seq: Sequence[Union[StrAny, str]]) -> StrAny:
    """
    Transforms a list of objects or strings [{K: {...}}, L, ...] -> {K: {...}, L: None, ...}
    """
    o: DictStrAny = {}
    for e in seq:
        if isinstance(e, dict):
            for k,v in e.items():
                if k in o:
                    raise KeyError(f"Cannot flatten with duplicate key {k}")
                o[k] = v
        else:
            key = str(e)
            if key in o:
                raise KeyError(f"Cannot flatten with duplicate key {k}")
            o[key] = None
    return o


def flatten_dicts_of_dicts(dicts: Mapping[str, Any]) -> Sequence[Any]:
    """
    Transform and object {K: {...}, L: {...}...} -> [{key:K, ....}, {key: L, ...}, ...]
    """
    o: List[Any] = []
    for k, v in dicts.items():
        if isinstance(v, list):
            # if v is a list then add "key" to each list element
            for lv in v:
                lv["key"] = k
        else:
            # add as "key" to dict
            v["key"] = k

        o.append(v)
    return o


def tuplify_list_of_dicts(dicts: Sequence[DictStrAny]) -> Sequence[DictStrAny]:
    """
    Transform dicts with single key into {"key": orig_key, "value": orig_value}
    """
    for d in dicts:
        if len(d) > 1:
            raise ValueError(f"Tuplify requires one key dicts {d}")
        if len(d) == 1:
            key = next(iter(d))
            # delete key first to avoid name clashes
            value = d[key]
            del d[key]
            d["key"] = key
            d["value"] = value

    return dicts


def filter_env_vars(envs: List[str]) -> StrStr:
    return {k.lower(): environ[k] for k in envs if k in environ}


def update_dict_with_prune(dest: DictStrAny, update: StrAny) -> None:
    for k, v in update.items():
        if v is not None:
            dest[k] = v
        elif k in dest:
            del dest[k]


def is_interactive() -> bool:
    import __main__ as main
    return not hasattr(main, '__file__')


@contextmanager
def custom_environ(env: StrStr) -> Iterator[None]:
    """Temporarily set environment variables inside the context manager and
    fully restore previous environment afterwards
    """
    original_env = {key: os.getenv(key) for key in env}
    os.environ.update(env)
    try:
        yield
    finally:
        for key, value in original_env.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value


def encoding_for_mode(mode: str) -> Optional[str]:
    if "b" in mode:
        return None
    else:
        return "utf-8"



from dlt.common.configuration.utils import TConfigSecret


class PostgresConfiguration:
    PG_DATABASE_NAME: str = None
    PG_SCHEMA_PREFIX: str = None
    PG_PASSWORD: TConfigSecret = None
    PG_USER: str = None
    PG_HOST: str = None
    PG_PORT: int = 5439
    PG_CONNECTION_TIMEOUT: int = 15

    @classmethod
    def check_integrity(cls) -> None:
        cls.PG_DATABASE_NAME = cls.PG_DATABASE_NAME.lower()
        cls.PG_SCHEMA_PREFIX = cls.PG_SCHEMA_PREFIX.lower()
        cls.PG_PASSWORD = TConfigSecret(cls.PG_PASSWORD.strip())


class PostgresProductionConfiguration(PostgresConfiguration):
    PG_DATABASE_NAME: str = None
    PG_SCHEMA_PREFIX: str = None
    PG_PASSWORD: TConfigSecret = None
    PG_USER: str = None
    PG_HOST: str = None



from collections.abc import Mapping as C_Mapping, Sequence as C_Sequence
from re import Pattern as _REPattern
from typing import Callable, Dict, Any, Literal, Mapping, Type, TypeVar, TypedDict, TYPE_CHECKING, Union, get_args, get_origin
if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath
    from typing import _TypedDict
    REPattern = _REPattern[str]
else:
    StrOrBytesPath = Any
    from typing import _TypedDictMeta as _TypedDict
    REPattern = _REPattern

DictStrAny = Dict[str, Any]
DictStrStr = Dict[str, str]
StrAny = Mapping[str, Any]  # immutable, covariant entity
StrStr = Mapping[str, str]  # immutable, covariant entity
StrStrStr = Mapping[str, Mapping[str, str]]  # immutable, covariant entity
TFun = TypeVar("TFun", bound=Callable[..., Any])


class TEvent(TypedDict, total=False):
    pass


class TTimestampEvent(TEvent, total=False):
    timestamp: float  # timestamp of event


def is_optional_type(t: Type[Any]) -> bool:
    # todo: use typing get_args and get_origin in python 3.8
    if hasattr(t, "__origin__"):
        return t.__origin__ is Union and type(None) in t.__args__
    return False


def extract_optional_type(t: Type[Any]) -> Any:
    return get_args(t)[0]


def is_literal_type(hint: Type[Any]) -> bool:
    return hasattr(hint, "__origin__") and hint.__origin__ is Literal


def is_typeddict(t: Any) -> bool:
    return isinstance(t, _TypedDict)


def is_list_generic_type(t: Any) -> bool:
    try:
        o = get_origin(t)
        return issubclass(o, list) or issubclass(o, C_Sequence)
    except Exception:
        return False


def is_dict_generic_type(t: Any) -> bool:
    try:
        o = get_origin(t)
        return issubclass(o, dict) or issubclass(o, C_Mapping)
    except Exception:
        return False



from contextlib import contextmanager
from pathlib import Path
from typing import Any, AnyStr, Dict, Iterator, List, Literal, Optional, Sequence, Tuple, Type
import google.cloud.bigquery as bigquery  # noqa: I250
from google.cloud.bigquery.dbapi import Connection as DbApiConnection
from google.cloud import exceptions as gcp_exceptions
from google.oauth2 import service_account
from google.api_core import exceptions as api_core_exceptions


from dlt.common import json, logger
from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.configuration import GcpClientConfiguration
from dlt.common.dataset_writers import TWriterType, escape_bigquery_identifier
from dlt.loaders.local_types import LoadJobStatus
from dlt.common.schema import TColumn, TDataType, Schema, TTableColumns

from dlt.loaders.client_base import JobClientBase, SqlClientBase, SqlJobClientBase, LoadJob, DBCursor, TJobClientCapabilities
from dlt.loaders.exceptions import LoadClientSchemaWillNotUpdate, LoadJobNotExistsException, LoadJobServerTerminalException, LoadUnknownTableException

SCT_TO_BQT: Dict[TDataType, str] = {
    "complex": "STRING",
    "text": "STRING",
    "double": "FLOAT64",
    "bool": "BOOLEAN",
    "timestamp": "TIMESTAMP",
    "bigint": "INTEGER",
    "binary": "BYTES",
    "decimal": f"NUMERIC({DEFAULT_NUMERIC_PRECISION},{DEFAULT_NUMERIC_SCALE})",
    "wei": "BIGNUMERIC"  # non parametrized should hold wei values
}

BQT_TO_SCT: Dict[str, TDataType] = {
    "STRING": "text",
    "FLOAT": "double",
    "BOOLEAN": "bool",
    "TIMESTAMP": "timestamp",
    "INTEGER": "bigint",
    "BYTES": "binary",
    "NUMERIC": "decimal",
    "BIGNUMERIC": "decimal"
}


class BigQuerySqlClient(SqlClientBase[bigquery.Client]):
    def __init__(self, schema_name: str, CREDENTIALS: Type[GcpClientConfiguration]) -> None:
        self._client: bigquery.Client = None
        self.C = CREDENTIALS
        super().__init__(schema_name)
        self.default_retry = bigquery.DEFAULT_RETRY.with_deadline(CREDENTIALS.TIMEOUT)
        self.default_query = bigquery.QueryJobConfig(default_dataset=self.fully_qualified_schema_name())

    def open_connection(self) -> None:
        credentials = service_account.Credentials.from_service_account_info(self.C.to_service_credentials())
        self._client = bigquery.Client(self.C.PROJECT_ID, credentials=credentials)

    def close_connection(self) -> None:
        if self._client:
            self._client.close()
            self._client = None

    @property
    def native_connection(self) -> bigquery.Client:
        return self._client

    def has_schema(self, schema_name: str = None) -> bool:
        dataset_name = self.fully_qualified_schema_name(schema_name)
        try:
            self._client.get_dataset(dataset_name, retry=self.default_retry, timeout=self.C.TIMEOUT)
            return True
        except gcp_exceptions.NotFound:
            return False

    def create_schema(self, schema_name: str = None) -> None:
        schema_name = self.fully_qualified_schema_name(schema_name)
        self._client.create_dataset(schema_name, exists_ok=False, retry=self.default_retry, timeout=self.C.TIMEOUT)

    def drop_schema(self, schema_name: str = None) -> None:
        schema_name = self.fully_qualified_schema_name(schema_name)
        self._client.delete_dataset(schema_name, not_found_ok=True, delete_contents=True, retry=self.default_retry, timeout=self.C.TIMEOUT)

    def execute_sql(self, sql: AnyStr, *args: Any, **kwargs: Any) -> Optional[Sequence[Sequence[Any]]]:
        logger.debug(f"Will execute query {sql}")  # type: ignore
        def_kwargs = {
            "job_config": self.default_query,
            "job_retry": self.default_retry,
            "timeout": self.C.TIMEOUT
            }
        kwargs = {**def_kwargs, **(kwargs or {})}
        results = self._client.query(sql, *args, **kwargs).result()
        if results:
            # consume and return all results
            return [list(r) for r in results]
        else:
            # no results were returned
            return None

    @contextmanager
    def execute_query(self, query: AnyStr,  *args: Any, **kwargs: Any) -> Iterator[DBCursor]:  # type: ignore
        conn: DbApiConnection = None
        def_kwargs = {
            "job_config": self.default_query
            }
        kwargs = {**def_kwargs, **(kwargs or {})}
        try:
            conn = DbApiConnection(client=self._client)
            curr = conn.cursor()
            curr.execute(query, *args, **kwargs)
            yield curr
        finally:
            if conn:
                # will also close all cursors
                conn.close()

    def fully_qualified_schema_name(self, schema_name: str = None) -> str:
        if schema_name:
            return f"{self.C.PROJECT_ID}.{schema_name}"
        else:
            return self._qualified_default_schema_name


class BigQueryLoadJob(LoadJob):
    def __init__(self, file_name: str, bq_load_job: bigquery.LoadJob, CONFIG: Type[GcpClientConfiguration]) -> None:
        self.bq_load_job = bq_load_job
        self.C = CONFIG
        self.default_retry = bigquery.DEFAULT_RETRY.with_deadline(CONFIG.TIMEOUT)
        super().__init__(file_name)

    def status(self) -> LoadJobStatus:
        # check server if done
        done = self.bq_load_job.done(retry=self.default_retry, timeout=self.C.TIMEOUT)
        if done:
            # rows processed
            if self.bq_load_job.output_rows is not None and self.bq_load_job.error_result is None:
                return "completed"
            else:
                return "failed"
        else:
            return "running"

    def file_name(self) -> str:
        return self._file_name

    def exception(self) -> str:
        exception: str = json.dumps({
            "error_result": self.bq_load_job.error_result,
            "errors": self.bq_load_job.errors,
            "job_start": self.bq_load_job.started,
            "job_end": self.bq_load_job.ended,
            "job_id": self.bq_load_job.job_id
        })
        return exception


class BigQueryClient(SqlJobClientBase):
    def __init__(self, schema: Schema, CONFIG: Type[GcpClientConfiguration]) -> None:
        sql_client = BigQuerySqlClient(schema.normalize_make_schema_name(CONFIG.DATASET, schema.schema_name), CONFIG)
        super().__init__(schema, sql_client)
        self.sql_client: BigQuerySqlClient = sql_client
        self.C = CONFIG

    def initialize_storage(self) -> None:
        if not self.sql_client.has_schema():
            self.sql_client.create_schema()

    def restore_file_load(self, file_path: str) -> LoadJob:
        try:
            return BigQueryLoadJob(
                JobClientBase.get_file_name_from_file_path(file_path),
                self._retrieve_load_job(file_path),
                self.C
            )
        except api_core_exceptions.NotFound:
            raise LoadJobNotExistsException(file_path)
        except (api_core_exceptions.BadRequest, api_core_exceptions.NotFound):
            raise LoadJobServerTerminalException(file_path)

    def start_file_load(self, table_name: str, file_path: str) -> LoadJob:
        # verify that table exists in the schema
        self._get_table_by_name(table_name, file_path)
        try:
            return BigQueryLoadJob(
                JobClientBase.get_file_name_from_file_path(file_path),
                self._create_load_job(table_name, file_path),
                self.C
            )
        except api_core_exceptions.NotFound:
            # google.api_core.exceptions.BadRequest - will not be processed ie bad job name
            raise LoadUnknownTableException(table_name, file_path)
        except (api_core_exceptions.BadRequest, api_core_exceptions.NotFound):
            # google.api_core.exceptions.NotFound: 404 - table not found
            raise LoadJobServerTerminalException(file_path)
        except api_core_exceptions.Conflict:
            # google.api_core.exceptions.Conflict: 409 PUT - already exists
            return self.restore_file_load(file_path)

    @property
    def capabilities(self) -> TJobClientCapabilities:
        return {
            "writer_type": supported_writer(self.C)
        }

    def _get_schema_version_from_storage(self) -> int:
        try:
            return super()._get_schema_version_from_storage()
        except api_core_exceptions.NotFound:
            # there's no table so there's no schema
            return 0

    def _build_schema_update_sql(self) -> List[str]:
        sql_updates = []
        for table_name in self.schema.schema_tables:
            exists, storage_table = self._get_storage_table(table_name)
            sql = self._get_table_update_sql(table_name, storage_table, exists)
            if sql:
                sql_updates.append(sql)
        return sql_updates

    def _get_table_update_sql(self, table_name: str, storage_table: TTableColumns, exists: bool) -> str:
        new_columns = self._create_table_update(table_name, storage_table)
        if len(new_columns) == 0:
            # no changes
            return None
        # build sql
        canonical_name = self.sql_client.fully_qualified_table_name(table_name)
        if not exists:
            # build CREATE
            sql = f"CREATE TABLE {canonical_name} (\n"
            sql += ",\n".join([self._get_column_def_sql(c) for c in new_columns])
            sql += ")"
        else:
            # build ALTER
            sql = f"ALTER TABLE {canonical_name}\n"
            sql += ",\n".join(["ADD COLUMN " + self._get_column_def_sql(c) for c in new_columns])
        # scan columns to get hints
        cluster_list = [escape_bigquery_identifier(c["name"]) for c in new_columns if c.get("cluster", False)]
        partition_list = [escape_bigquery_identifier(c["name"]) for c in new_columns if c.get("partition", False)]
        # partition by must be added first
        if len(partition_list) > 0:
            if exists:
                raise LoadClientSchemaWillNotUpdate(canonical_name, partition_list, "Partition requested after table was created")
            elif len(partition_list) > 1:
                raise LoadClientSchemaWillNotUpdate(canonical_name, partition_list, "Partition requested for more than one column")
            else:
                sql += f"\nPARTITION BY DATE({partition_list[0]})"
        if len(cluster_list) > 0:
            if exists:
                raise LoadClientSchemaWillNotUpdate(canonical_name, cluster_list, "Clustering requested after table was created")
            else:
                sql += "\nCLUSTER BY " + ",".join(cluster_list)

        return sql

    def _get_column_def_sql(self, c: TColumn) -> str:
        name = escape_bigquery_identifier(c["name"])
        return f"{name} {self._sc_t_to_bq_t(c['data_type'])} {self._gen_not_null(c['nullable'])}"

    def _get_storage_table(self, table_name: str) -> Tuple[bool, TTableColumns]:
        schema_table: TTableColumns = {}
        try:
            table = self.sql_client.native_connection.get_table(
                self.sql_client.fully_qualified_table_name(table_name), retry=self.sql_client.default_retry, timeout=self.C.TIMEOUT
            )
            partition_field = table.time_partitioning.field if table.time_partitioning else None
            for c in table.schema:
                schema_c: TColumn = {
                    "name": c.name,
                    "nullable": c.is_nullable,
                    "data_type": self._bq_t_to_sc_t(c.field_type, c.precision, c.scale),
                    "unique": False,
                    "sort": False,
                    "primary_key": False,
                    "foreign_key": False,
                    "cluster": c.name in (table.clustering_fields or []),
                    "partition": c.name == partition_field
                }
                schema_table[c.name] = schema_c
            return True, schema_table
        except gcp_exceptions.NotFound:
            return False, schema_table

    def _create_load_job(self, table_name: str, file_path: str) -> bigquery.LoadJob:
        job_id = BigQueryClient._get_job_id_from_file_path(file_path)
        job_config = bigquery.LoadJobConfig(
            autodetect=False,
            write_disposition=bigquery.WriteDisposition.WRITE_APPEND,
            create_disposition=bigquery.CreateDisposition.CREATE_NEVER,
            source_format=bigquery.SourceFormat.NEWLINE_DELIMITED_JSON,
            ignore_unknown_values=False,
            max_bad_records=0,

            )
        with open(file_path, "rb") as f:
            return self.sql_client.native_connection.load_table_from_file(f,
                                                     self.sql_client.fully_qualified_table_name(table_name),
                                                     job_id=job_id,
                                                     job_config=job_config,
                                                     timeout=self.C.TIMEOUT
                                                    )

    def _retrieve_load_job(self, file_path: str) -> bigquery.LoadJob:
        job_id = BigQueryClient._get_job_id_from_file_path(file_path)
        return self.sql_client.native_connection.get_job(job_id)

    @staticmethod
    def _get_job_id_from_file_path(file_path: str) -> str:
        return Path(file_path).name.replace(".", "_")

    @staticmethod
    def _gen_not_null(v: bool) -> str:
        return "NOT NULL" if not v else ""

    @staticmethod
    def _sc_t_to_bq_t(sc_t: TDataType) -> str:
        return SCT_TO_BQT[sc_t]

    @staticmethod
    def _bq_t_to_sc_t(bq_t: str, precision: Optional[int], scale: Optional[int]) -> TDataType:
        if bq_t == "BIGNUMERIC":
            if precision is None:  # biggest numeric possible
                return "wei"
        return BQT_TO_SCT.get(bq_t, "text")


def make_client(schema: Schema, C: Type[GcpClientConfiguration]) -> BigQueryClient:
    return BigQueryClient(schema, C)


def supported_writer(C: Type[GcpClientConfiguration]) -> TWriterType:
    return "jsonl"



import decimal # noqa: I251
from contextlib import contextmanager
from typing import Iterator
from decimal import ROUND_HALF_UP, Decimal, DefaultContext, DivisionByZero, InvalidOperation, localcontext, Context, ConversionSyntax  # noqa: I251


DEFAULT_NUMERIC_PRECISION = 38
DEFAULT_NUMERIC_SCALE = 9

NUMERIC_DEFAULT_QUANTIZER = Decimal("1." + "0" * DEFAULT_NUMERIC_SCALE)

DefaultContext.rounding = ROUND_HALF_UP
# use small caps for exponent
DefaultContext.capitals = 0
# use 128 bit precision which is default in most databases
DefaultContext.prec = DEFAULT_NUMERIC_PRECISION
# prevent NaN to be returned
DefaultContext.traps[InvalidOperation] = True
# prevent Inf to be returned
DefaultContext.traps[DivisionByZero] = True
decimal.setcontext(DefaultContext)


@contextmanager
def numeric_default_context(precision: int = DEFAULT_NUMERIC_PRECISION) -> Iterator[Context]:
    with localcontext() as c:
        c.prec = precision
        yield c


def numeric_default_quantize(v: Decimal) -> Decimal:
    if v == 0:
        return v
    return v.quantize(NUMERIC_DEFAULT_QUANTIZER)



from importlib import import_module
import yaml
from copy import copy
from typing import Dict, List, Mapping, Optional, Sequence, Tuple, Any, cast

from dlt.common.typing import DictStrAny, StrAny, REPattern
from dlt.common.normalizers.names import TNormalizeBreakPath, TNormalizeMakePath, TNormalizeNameFunc
from dlt.common.normalizers.json import TNormalizeJSONFunc
from dlt.common.schema.typing import TNormalizersConfig, TPartialTable, TSchemaSettings, TSimpleRegex, TStoredSchema, TSchemaTables, TTable, TTableColumns, TColumn, TColumnProp, TDataType, THintType
from dlt.common.schema import utils
from dlt.common.schema.exceptions import (CannotCoerceColumnException, CannotCoerceNullException, InvalidSchemaName,
                                          ParentTableNotFoundException, SchemaCorruptedException, TablePropertiesClashException)
from dlt.common.validation import validate_dict


class Schema:

    VERSION_TABLE_NAME = "_dlt_version"
    VERSION_COLUMN_NAME = "version"
    LOADS_TABLE_NAME = "_dlt_loads"
    ENGINE_VERSION = 3

    def __init__(self, name: str, normalizers: TNormalizersConfig = None) -> None:
        # verify schema name
        if name != utils.normalize_schema_name(name):
            raise InvalidSchemaName(name, utils.normalize_schema_name(name))
        self._schema_tables: TSchemaTables = {}
        self._schema_name: str = name
        self._version = 1
        # schema settings to hold default hints, preferred types and other settings
        self._settings: TSchemaSettings = {}

        # list of preferred types: map regex on columns into types
        self._compiled_preferred_types: List[Tuple[REPattern, TDataType]] = []
        # compiled default hints
        self._compiled_hints: Dict[THintType, Sequence[REPattern]] = {}
        # compiled exclude filters per table
        self._compiled_excludes: Dict[str, Sequence[REPattern]] = {}
        # compiled include filters per table
        self._compiled_includes: Dict[str, Sequence[REPattern]] = {}

        # normalizers config
        self._normalizers_config: TNormalizersConfig = normalizers
        # name normalization functions
        self.normalize_table_name: TNormalizeNameFunc = None
        self.normalize_column_name: TNormalizeNameFunc = None
        self.normalize_schema_name: TNormalizeNameFunc = None
        self.normalize_make_schema_name: TNormalizeMakePath = None
        self.normalize_make_path: TNormalizeMakePath = None
        self.normalize_break_path: TNormalizeBreakPath = None
        # json normalization function
        self.normalize_json: TNormalizeJSONFunc = None

        # add version tables
        self._add_standard_tables()
        # add standard hints
        self._add_standard_hints()
        # configure normalizers, including custom config if present
        self._configure_normalizers()
        # compile all known regexes
        self._compile_regexes()

    @classmethod
    def from_dict(cls, d: DictStrAny) -> "Schema":
        # upgrade engine if needed
        stored_schema = utils.upgrade_engine_version(d, d["engine_version"], cls.ENGINE_VERSION)
        # verify schema
        utils.validate_stored_schema(stored_schema)
        # add defaults
        utils.apply_defaults(stored_schema)

        # create new instance from dict
        self: Schema = cls(stored_schema["name"], normalizers=stored_schema.get("normalizers", None))
        self._schema_tables = stored_schema.get("tables", {})
        # TODO: generate difference if STANDARD SCHEMAS are different than those and increase schema version
        if Schema.VERSION_TABLE_NAME not in self._schema_tables:
            raise SchemaCorruptedException(f"Schema must contain table {Schema.VERSION_TABLE_NAME}")
        if Schema.LOADS_TABLE_NAME not in self._schema_tables:
            raise SchemaCorruptedException(f"Schema must contain table {Schema.LOADS_TABLE_NAME}")
        self._version = stored_schema["version"]
        self._settings = stored_schema.get("settings", {})
        # compile regexes
        self._compile_regexes()

        return self

    def filter_row(self, table_name: str, row: StrAny) -> StrAny:
        # exclude row elements according to the rules in `filter` elements of the table
        # include rules have precedence and are used to make exceptions to exclude rules
        # the procedure will apply rules from the table_name and it's all parent tables up until root
        # parent tables are computed by `normalize_break_path` function so they do not need to exist in the schema
        # note: the above is not very clean. the `parent` element of each table should be used but as the rules
        #  are typically used to prevent not only table fields but whole tables from being created it is not possible

        def _exclude(path: str, excludes: Sequence[REPattern], includes: Sequence[REPattern]) -> bool:
            is_included = False
            is_excluded = any(exclude.search(path) for exclude in excludes)
            if is_excluded:
                # we may have exception if explicitly included
                is_included = any(include.search(path) for include in includes)
            return is_excluded and not is_included

        # break table name in components
        branch = self.normalize_break_path(table_name)

        # check if any of the rows is excluded by rules in any of the tables
        for i in range(len(branch), 0, -1):  # stop is exclusive in `range`
            # start at the top level table
            c_t = self.normalize_make_path(*branch[:i])
            excludes = self._compiled_excludes.get(c_t)
            # only if there's possibility to exclude, continue
            if excludes:
                includes = self._compiled_includes.get(c_t, [])
                for field_name in list(row.keys()):
                    path = self.normalize_make_path(*branch[i:], field_name)
                    if _exclude(path, excludes, includes):
                        # TODO: copy to new instance
                        del row[field_name]  # type: ignore
            # if row is empty, do not process further
            if not row:
                break
        return row

    def coerce_row(self, table_name: str, parent_table: str, row: StrAny) -> Tuple[StrAny, TPartialTable]:
        # get existing or create a new table
        table = self._schema_tables.get(table_name, utils.new_table(table_name, parent_table))
        table_columns = table["columns"]

        partial_table: TPartialTable = None
        new_row: DictStrAny = {}
        for col_name, v in row.items():
            # skip None values, we should infer the types later
            if v is None:
                # just check if column is nullable if exists
                self._coerce_null_value(table_columns, table_name, col_name)
            else:
                new_col_name, new_col_def, new_v = self._coerce_non_null_value(table_columns, table_name, col_name, v)
                new_row[new_col_name] = new_v
                if new_col_def:
                    if not partial_table:
                        partial_table = copy(table)
                        partial_table["columns"] = {}
                    partial_table["columns"][new_col_name] = new_col_def

        return new_row, partial_table

    def update_schema(self, partial_table: TPartialTable) -> None:
        table_name = partial_table["name"]
        parent_table_name = partial_table.get("parent")
        # check if parent table present
        if parent_table_name is not None:
            if self._schema_tables.get(parent_table_name) is None:
                raise ParentTableNotFoundException(
                    table_name, parent_table_name,
                    f" This may be due to misconfigured excludes filter that fully deletes content of the {parent_table_name}. Add includes that will preserve the parent table."
                    )
        table = self._schema_tables.get(table_name)
        if table is None:
            # add the whole new table to SchemaTables
            self._schema_tables[table_name] = partial_table
        else:
            # check if table properties can be merged
            if table.get("parent") != partial_table.get("parent"):
                raise TablePropertiesClashException(table_name, "parent", table.get("parent"), partial_table.get("parent"))
            if table.get("write_disposition") != partial_table.get("write_disposition"):
                raise TablePropertiesClashException(table_name, "write_disposition", table.get("write_disposition"), partial_table.get("write_disposition"))
            # add several columns to existing table
            table_columns = table["columns"]
            for column in partial_table["columns"].values():
                column_name = column["name"]
                if column_name in table_columns:
                    # we do not support changing existing columns
                    if not utils.compare_columns(table_columns[column_name], column):
                        # attempt to update to incompatible columns
                        raise CannotCoerceColumnException(table_name, column_name, table_columns[column_name]["data_type"], column["data_type"], None)
                else:
                    table_columns[column_name] = column
        # bump schema version
        self._version += 1

    def filter_row_with_hint(self, table_name: str, hint_type: THintType, row: StrAny) -> StrAny:
        rv_row: DictStrAny = {}
        column_prop: TColumnProp = utils.hint_to_column_prop(hint_type)
        try:
            table = self.get_table_columns(table_name)
            for column_name in table:
                if column_name in row:
                    hint_value = table[column_name][column_prop]
                    if (hint_value and column_prop != "nullable") or (column_prop == "nullable" and not hint_value):
                        rv_row[column_name] = row[column_name]
        except KeyError:
            for k, v in row.items():
                if self._infer_hint(hint_type, v, k):
                    rv_row[k] = v

        # dicts are ordered and we will return the rows with hints in the same order as they appear in the columns
        return rv_row

    def merge_hints(self, new_hints: Mapping[THintType, Sequence[TSimpleRegex]]) -> None:
        # validate regexes
        validate_dict(TSchemaSettings, {"default_hints": new_hints}, ".", validator_f=utils.simple_regex_validator)
        # prepare hints to be added
        default_hints = self._settings.setdefault("default_hints", {})
        # add `new_hints` to existing hints
        for h, l in new_hints.items():
            if h in default_hints:
                # merge if hint of this type exist
                default_hints[h] = list(set(default_hints[h] + list(l)))
            else:
                # set new hint type
                default_hints[h] = l  # type: ignore
        self._compile_regexes()

    def get_schema_update_for(self, table_name: str, t: TTableColumns) -> List[TColumn]:
        # gets new columns to be added to "t" to bring up to date with stored schema
        diff_c: List[TColumn] = []
        s_t = self.get_table_columns(table_name)
        for c in s_t.values():
            if c["name"] not in t:
                diff_c.append(c)
        return diff_c

    def get_table(self, table_name: str) -> TTable:
        return self._schema_tables[table_name]

    def get_table_columns(self, table_name: str) -> TTableColumns:
        return self._schema_tables[table_name]["columns"]

    def all_tables(self, with_dlt_tables: bool = False) -> List[TTable]:
        return [t for t in self._schema_tables.values() if not t["name"].startswith("_dlt") or with_dlt_tables]

    def get_preferred_type(self, col_name: str) -> Optional[TDataType]:
        return next((m[1] for m in self._compiled_preferred_types if m[0].search(col_name)), None)

    def to_dict(self, remove_defaults: bool = False) -> TStoredSchema:
        stored_schema: TStoredSchema = {
            "version": self._version,
            "engine_version": Schema.ENGINE_VERSION,
            "name": self._schema_name,
            "tables": self._schema_tables,
            "settings": self._settings,
            "normalizers": self._normalizers_config
        }
        if remove_defaults:
            utils.remove_defaults(stored_schema)
        return stored_schema

    @property
    def schema_version(self) -> int:
        return self._version

    @property
    def schema_name(self) -> str:
        return self._schema_name

    @property
    def schema_tables(self) -> TSchemaTables:
        return self._schema_tables

    @property
    def schema_settings(self) -> TSchemaSettings:
        return self._settings

    def as_yaml(self, remove_defaults: bool = False) -> str:
        d = self.to_dict(remove_defaults=remove_defaults)
        return cast(str, yaml.dump(d, allow_unicode=True, default_flow_style=False, sort_keys=False))

    def _infer_column(self, k: str, v: Any) -> TColumn:
        return TColumn(
            name=k,
            data_type=self._map_value_to_column_type(v, k),
            nullable=not self._infer_hint("not_null", v, k),
            partition=self._infer_hint("partition", v, k),
            cluster=self._infer_hint("cluster", v, k),
            sort=self._infer_hint("sort", v, k),
            unique=self._infer_hint("unique", v, k),
            primary_key=self._infer_hint("primary_key", v, k),
            foreign_key=self._infer_hint("foreign_key", v, k)
        )

    def _coerce_null_value(self, table_columns: TTableColumns, table_name: str, col_name: str) -> None:
        if col_name in table_columns:
            existing_column = table_columns[col_name]
            if not existing_column["nullable"]:
                raise CannotCoerceNullException(table_name, col_name)

    def _coerce_non_null_value(self, table_columns: TTableColumns, table_name: str, col_name: str, v: Any) -> Tuple[str, TColumn, Any]:
        new_column: TColumn = None
        variant_col_name = col_name

        if col_name in table_columns:
            existing_column = table_columns[col_name]
            # existing columns cannot be changed so we must update row
            py_data_type = utils.py_type_to_sc_type(type(v))
            # first try to coerce existing value into destination type
            try:
                rv = utils.coerce_type(existing_column["data_type"], py_data_type, v)
            except (ValueError, SyntaxError):
                # if that does not work we must create variant extension to the table
                variant_col_name = f"{col_name}_v_{py_data_type}"
                # if variant exists check type, coercions are not required
                if variant_col_name in table_columns:
                    if table_columns[variant_col_name]["data_type"] != py_data_type:
                        raise CannotCoerceColumnException(table_name, variant_col_name, table_columns[variant_col_name]["data_type"], py_data_type, v)
                else:
                    # add new column
                    new_column = self._infer_column(variant_col_name, v)
                    # must have variant type, not preferred or coerced type
                    new_column["data_type"] = py_data_type
                # coerce even if types are the same (complex case)
                rv = utils.coerce_type(py_data_type, py_data_type, v)
        else:
            # infer new column
            new_column = self._infer_column(col_name, v)
            # and coerce type if inference changed the python type
            py_type = utils.py_type_to_sc_type(type(v))
            rv = utils.coerce_type(new_column["data_type"], py_type, v)

        return variant_col_name, new_column, rv

    def _map_value_to_column_type(self, v: Any, k: str) -> TDataType:
        tv = type(v)
        # try to autodetect data type
        mapped_type = utils.autodetect_sc_type(self._normalizers_config.get("detections"), tv, v)
        # if not try standard type mapping
        if mapped_type is None:
            mapped_type = utils.py_type_to_sc_type(tv)
        # get preferred type based on column name
        preferred_type = self.get_preferred_type(k)
        # try to match python type to preferred
        if preferred_type:
            # try to coerce to destination type
            try:
                utils.coerce_type(preferred_type, mapped_type, v)
                # coercion possible so preferred type may be used
                mapped_type = preferred_type
            except ValueError:
                # coercion not possible
                pass
        return mapped_type

    def _infer_hint(self, hint_type: THintType, _: Any, k: str) -> bool:
        if hint_type in self._compiled_hints:
            return any(h.search(k) for h in self._compiled_hints[hint_type])
        else:
            return False

    def _add_standard_tables(self) -> None:
        self._schema_tables[Schema.VERSION_TABLE_NAME] = utils.version_table()
        self._schema_tables[Schema.LOADS_TABLE_NAME] = utils.load_table()

    def _add_standard_hints(self) -> None:
        default_hints = utils.standard_hints()
        if default_hints:
            self._settings["default_hints"] = default_hints

    def _configure_normalizers(self) -> None:
        if not self._normalizers_config:
            # create default normalizer config
            self._normalizers_config = utils.default_normalizers()
        # import desired modules
        naming_module = import_module(self._normalizers_config["names"])
        json_module = import_module(self._normalizers_config["json"]["module"])
        # name normalization functions
        self.normalize_table_name = naming_module.normalize_table_name
        self.normalize_column_name = naming_module.normalize_column_name
        self.normalize_schema_name = utils.normalize_schema_name
        self.normalize_make_schema_name = naming_module.normalize_make_schema_name
        self.normalize_make_path = naming_module.normalize_make_path
        self.normalize_break_path = naming_module.normalize_break_path
        # json normalization function
        self.normalize_json = json_module.normalize
        json_module.extend_schema(self)

    def _compile_regexes(self) -> None:
        if self._settings:
            for pattern, dt in self._settings.get("preferred_types", {}).items():
                # add tuples to be searched in coercions
                self._compiled_preferred_types.append((utils.compile_simple_regex(pattern), dt))
            for hint_name, hint_list in self._settings.get("default_hints", {}).items():
                # compile hints which are column matching regexes
                self._compiled_hints[hint_name] = list(map(lambda hint: utils.compile_simple_regex(hint), hint_list))
        if self._schema_tables:
            for table in self._schema_tables.values():
                if "filters" in table:
                    if "excludes" in table["filters"]:
                        self._compiled_excludes[table["name"]] = list(map(lambda exclude: utils.compile_simple_regex(exclude), table["filters"]["excludes"]))
                    if "includes" in table["filters"]:
                        self._compiled_includes[table["name"]] = list(map(lambda exclude: utils.compile_simple_regex(exclude), table["filters"]["includes"]))



# ─────── DLT ───────────────
# ───▄▄█████████▄────────────
# ──███████▄██▀▀─────────────
# ─▐████████── JSON ── JSON──
# ──█████████▄▄──────────────
# ───▀██████████▀────────────

from typing import Sequence

from dlt.common.typing import DictStrAny
from dlt.common import json
from dlt.common.schema import Schema
from dlt.pipeline import Pipeline, GCPPipelineCredentials

# the load schema will be named {pipeline_name}_{source_name}
# this allows you to easily consume multiple environments/instances of the same source
schema_prefix = 'library_documents'
schema_source_suffix = 'prod'

# you authenticate by passing a credential, such as RDBMS host/user/port/pass or gcp service account json.
gcp_credential_json_file_path = "/Users/adrian/PycharmProjects/sv/dlt/temp/scalevector-1235ac340b0b.json"

# get some data and give a name to the parent table.
# If you have a stream with different document types, pass the doc type into the table name to separate the stream.
# In our example we have a list of 2 books

data_file_path = "/Users/adrian/PycharmProjects/sv/dlt/examples/data/demo_example.json"
parent_table = 'books'

data_schema = None
data_schema_file_path = "/Users/adrian/PycharmProjects/sv/dlt/examples/schemas/inferred_demo_schema.yml"

# or pass a saved data schema file
# f = open(data_schema_file_path, "r", encoding="utf-8")
# data_schema = f.read()
#
#

"""
to get your data in a flat table, you need to join the sub-json back to the parent
example sql:

SELECT
    b.* EXCEPT(_dlt_load_id, _dlt_id, _dlt_root_id),
    bc.* EXCEPT(_dlt_parent_id, _dlt_id, _dlt_root_id),
FROM `scalevector.library_documents_prod.books` as b
left join `scalevector.library_documents_prod.books__category` as bc
on b._dlt_id = bc._dlt_parent_id

SQL result rows:
{  "isbn": "123-456-222",  "author__lastname": "Panda",  "author__firstname": "Jane",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "The Ultimate Database Study Guide",  "value": "Non-Fiction",  "_dlt_list_idx": "0"}
{  "isbn": "123-456-222",  "author__lastname": "Panda",  "author__firstname": "Jane",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "The Ultimate Database Study Guide",  "value": "Technology",  "_dlt_list_idx": "1"}
{  "isbn": "123-456-789",  "author__lastname": "Jayson",  "author__firstname": "Joe",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "Json for big data",  "value": "SF",  "_dlt_list_idx": "0"}
{  "isbn": "123-456-789",  "author__lastname": "Jayson",  "author__firstname": "Joe",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "Json for big data",  "value": "Horror",  "_dlt_list_idx": "1"}
{  "isbn": "123-456-789",  "author__lastname": "Jayson",  "author__firstname": "Joe",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "Json for big data",  "value": "Dystopia",  "_dlt_list_idx": "2"}
"""  # noqa: E501


if __name__ == "__main__":

    # loading and error handling below:

    def get_json_file_data(path: str) -> Sequence[DictStrAny]:
        with open(path, "r", encoding="utf-8") as f:
            data = json.load(f)
        return data  # type: ignore

    data = get_json_file_data(data_file_path)

    # this is example of extracting from iterator that may be a
    # - jsonl file you read line by line
    # - a list of dicts
    # - iterator over panda frame https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_dict.html
    # - any iterator with map function
    # - any iterator whatsoever ie getting data from REST API

    # extracting from iterator always happens in single worker thread and iterator is directly returning data so there's no retry mechanism if anything fails

    # working BQ creds
    # credentials = Pipeline.load_gcp_credentials("_secrets/project1234_service.json", "gamma_guild")

    # working redshift creds, you can pass password as last parameter or via PG_PASSWORD env variable ie.
    # LOG_LEVEL=INFO PG_PASSWORD=.... python examples/discord_iterator.py
    credentials = GCPPipelineCredentials.from_services_file(gcp_credential_json_file_path, schema_prefix)

    pipeline = Pipeline(schema_source_suffix)

    schema: Schema = None
    # uncomment to use already modified schema, but the auto-inferred schema will also work nicely
    # TODO: schema below needs is not yet finished, more exploration needed
    # schema = Pipeline.load_schema_from_file("examples/schemas/discord_schema.yml")
    pipeline.create_pipeline(credentials, schema=schema)
    # the pipeline created a working directory. you can always attach the pipeline back by just providing the dir to
    # Pipeline::restore_pipeline
    print(pipeline.root_path)

    # and extract it
    pipeline.extract(iter(data), table_name=parent_table)

    # now create loading packages and infer the schema
    pipeline.unpack()

    schema = pipeline.get_default_schema()
    schema_yaml = schema.as_yaml()
    f = open(data_schema_file_path, "a", encoding="utf-8")
    f.write(schema_yaml)
    f.close()
    # pipeline.save_schema_to_file(data_schema_file_path, schema)

    # show loads, each load contains a copy of the schema that corresponds to the data inside
    # and a set of directories for job states (new -> in progress -> failed|completed)
    new_loads = pipeline.list_unpacked_loads()
    print(new_loads)

    # load packages
    pipeline.load()

    # should be empty
    new_loads = pipeline.list_unpacked_loads()
    print(new_loads)

    # now enumerate all complete loads if we have any failed packages
    # complete but failed job will not raise any exceptions
    completed_loads = pipeline.list_completed_loads()
    # print(completed_loads)
    for load_id in completed_loads:
        print(f"Checking failed jobs in {load_id}")
        for job, failed_message in pipeline.list_failed_jobs(load_id):
            print(f"JOB: {job}\nMSG: {failed_message}")



import base64
from dlt.common import json
from dlt.common.utils import uniq_id
from dlt.pipeline import Pipeline, GCPPipelineCredentials

# 1. configuration: name your schema, table, pass credentials
schema_prefix = 'demo_' + uniq_id()[:4]
schema_name = 'example'
parent_table = 'my_json_doc'
# gcp_credential_json_file_path = "/Users/adrian/PycharmProjects/sv/dlt/temp/scalevector-1235ac340b0b.json"
gcp_credentials_json = {
    "type": "service_account",
    "project_id": "zinc-mantra-353207",
    "private_key": "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",  # noqa
    "client_email": "data-load-tool-public-demo@zinc-mantra-353207.iam.gserviceaccount.com",
}

# we do not want to have this key verbatim in repo so we decode it here
gcp_credentials_json["private_key"] = bytes([_a ^ _b for _a, _b in zip(base64.b64decode(gcp_credentials_json["private_key"]), b"quickstart-sv"*150)]).decode("utf-8")

# if you re-use an edited schema, then uncomment this part, so you can save it to file
schema_file_path = "examples/schemas/quickstart.yml"


# 2. Create a pipeline
credential = GCPPipelineCredentials.from_services_dict(gcp_credentials_json, schema_prefix)
pipeline = Pipeline(schema_name)
pipeline.create_pipeline(credential)

# If you want to re-use a curated schema, uncomment the below
# schema = Pipeline.load_schema_from_file(schema_file_path)
# pipeline.create_pipeline(credential, schema=schema)


# 3. Pass the data to the pipeline and give it a table name. Optionally unpack and handle schema.

# with open("experiments/metabase_data.json", "r") as f:
#     rows = json.load(f)
rows = [{"name": "Ana", "age": 30, "id": 456, "children": [{"name": "Bill", "id": 625},
                                                           {"name": "Elli", "id": 591}
                                                           ]},

        {"name": "Bob", "age": 30, "id": 455, "children": [{"name": "Bill", "id": 625},
                                                           {"name": "Dave", "id": 621}
                                                           ]}
        ]

pipeline.extract(iter(rows), table_name=parent_table)
# tell the pipeline to un-nest the json into a relational structure
pipeline.unpack()

# If you want to save the schema to curate it and re-use it, uncomment the below
schema = pipeline.get_default_schema()
schema_yaml = schema.as_yaml(remove_defaults=True)
print(schema_yaml)
# f = open(data_schema_file_path, "a", encoding="utf-8")
# f.write(schema_yaml)
# f.close()

# 4. Load
pipeline.load()

# 5. Optional error handling - print, raise or handle.

# now enumerate all complete loads if we have any failed packages
# complete but failed job will not raise any exceptions
completed_loads = pipeline.list_completed_loads()
# print(completed_loads)
for load_id in completed_loads:
    print(f"Checking failed jobs in {load_id}")
    for job, failed_message in pipeline.list_failed_jobs(load_id):
        print(f"JOB: {job}\nMSG: {failed_message}")

# now you can use your data

# Tables created:

with pipeline.sql_client() as c:
    query = "SELECT * FROM my_json_doc"
    df = c.execute_sql(query)
    print(query)
    print(list(df))
    print()
    # {  "name": "Ana",  "age": "30",  "id": "456",  "_dlt_load_id": "1654787700.406905",  "_dlt_id": "5b018c1ba3364279a0ca1a231fbd8d90"}
    # {  "name": "Bob",  "age": "30",  "id": "455",  "_dlt_load_id": "1654787700.406905",  "_dlt_id": "afc8506472a14a529bf3e6ebba3e0a9e"}

    query = "SELECT * FROM my_json_doc__children LIMIT 1000"
    df = c.execute_sql(query)
    print(query)
    print(list(df))
    print()
    # {"name": "Bill", "id": "625", "_dlt_parent_id": "5b018c1ba3364279a0ca1a231fbd8d90", "_dlt_list_idx": "0", "_dlt_root_id": "5b018c1ba3364279a0ca1a231fbd8d90",
    #   "_dlt_id": "7993452627a98814cc7091f2c51faf5c"}
    # {"name": "Bill", "id": "625", "_dlt_parent_id": "afc8506472a14a529bf3e6ebba3e0a9e", "_dlt_list_idx": "0", "_dlt_root_id": "afc8506472a14a529bf3e6ebba3e0a9e",
    #   "_dlt_id": "9a2fd144227e70e3aa09467e2358f934"}
    # {"name": "Dave", "id": "621", "_dlt_parent_id": "afc8506472a14a529bf3e6ebba3e0a9e", "_dlt_list_idx": "1", "_dlt_root_id": "afc8506472a14a529bf3e6ebba3e0a9e",
    #   "_dlt_id": "28002ed6792470ea8caf2d6b6393b4f9"}
    # {"name": "Elli", "id": "591", "_dlt_parent_id": "5b018c1ba3364279a0ca1a231fbd8d90", "_dlt_list_idx": "1", "_dlt_root_id": "5b018c1ba3364279a0ca1a231fbd8d90",
    #   "_dlt_id": "d18172353fba1a492c739a7789a786cf"}

    # and we can join them via auto generated keys
    query = """
        select p.name, p.age, p.id as parent_id,
            c.name as child_name, c.id as child_id, c._dlt_list_idx as child_order_in_list
        from my_json_doc as p
        left join my_json_doc__children  as c
            on p._dlt_id = c._dlt_parent_id
    """
    df = c.execute_sql(query)
    print(query)
    print(list(df))
    print()
    # {  "name": "Ana",  "age": "30",  "parent_id": "456",  "child_name": "Bill",  "child_id": "625",  "child_order_in_list": "0"}
    # {  "name": "Ana",  "age": "30",  "parent_id": "456",  "child_name": "Elli",  "child_id": "591",  "child_order_in_list": "1"}
    # {  "name": "Bob",  "age": "30",  "parent_id": "455",  "child_name": "Bill",  "child_id": "625",  "child_order_in_list": "0"}
    # {  "name": "Bob",  "age": "30",  "parent_id": "455",  "child_name": "Dave",  "child_id": "621",  "child_order_in_list": "1"}



from dlt.common.schema import Schema

from dlt.pipeline import Pipeline


def discover_schema() -> Schema:
    # makes all 'timestamp' fields of type timestamps and filters out response selectors
    return Pipeline.load_schema_from_file("examples/schemas/rasa_schema.yml")



from dlt.common.schema.typing import TSchemaUpdate, TStoredSchema, TTableColumns, TDataType, THintType, TColumn, TColumnBase  # noqa: F401
from dlt.common.schema.typing import COLUMN_HINTS  # noqa: F401
from dlt.common.schema.schema import Schema  # noqa: F401
from dlt.common.schema.utils import normalize_schema_name, add_missing_hints  # noqa: F401



import os
import tempfile
from typing import Iterator, TypedDict, cast, Union

from dlt.common import json
from dlt.common.runners.venv import Venv
from dlt.common.sources import with_table_name
from dlt.common.typing import DictStrAny, StrAny, StrOrBytesPath

from examples.sources.stdout import get_source as get_singer_pipe

# from dlt.pipeline.exceptions import MissingDependencyException
FilePathOrDict = Union[StrAny, StrOrBytesPath]


class SingerMessage(TypedDict):
    type: str  # noqa: A003


class SingerRecord(SingerMessage):
    record: DictStrAny
    stream: str


class SingerState(SingerMessage):
    value: DictStrAny

# try:
#     from singer import parse_message_from_obj, Message, RecordMessage, StateMessage
# except ImportError:
#     raise MissingDependencyException("Singer Source", ["python-dlt-singer"], "Singer runtime compatible with DLT")


# pip install ../singer/singer-python
# https://github.com/datamill-co/singer-runner/tree/master/singer_runner
# https://techgaun.github.io/active-forks/index.html#singer-io/singer-python
def get_source_from_stream(singer_messages: Iterator[SingerMessage], state: DictStrAny = None) -> Iterator[DictStrAny]:
    last_state = {}
    for msg in singer_messages:
        if msg["type"] == "RECORD":
            # yield record
            msg = cast(SingerRecord, msg)
            yield with_table_name(msg["record"], msg["stream"])
        if msg["type"] == "STATE":
            msg = cast(SingerState, msg)
            last_state = msg["value"]
    if state:
        state["singer"] = last_state


def get_source(venv: Venv, tap_name: str, config_file: FilePathOrDict, catalog_file: FilePathOrDict, state: DictStrAny = None) -> Iterator[DictStrAny]:
    # write config dictionary to temp file in virtual environment if passed as dict
    config_file_path: StrOrBytesPath = None
    if type(config_file) is dict:
        fd, tmp_name = tempfile.mkstemp(dir=venv.context.env_dir)
        with os.fdopen(fd, "w") as f:
            json.dump(config_file, f)
            config_file_path = tmp_name
    else:
        config_file_path = config_file  # type: ignore

    # process catalog like config
    catalog_file_path: StrOrBytesPath = None
    if type(catalog_file) is dict:
        fd, tmp_name = tempfile.mkstemp(dir=venv.context.env_dir)
        with os.fdopen(fd, "w") as f:
            json.dump(catalog_file, f)
            catalog_file_path = tmp_name
    else:
        catalog_file_path = catalog_file  # type: ignore

    pipe_iterator = get_singer_pipe(venv,
                                    tap_name,
                                    "--config",
                                    os.path.abspath(config_file_path),
                                    "--catalog",
                                    os.path.abspath(catalog_file_path)
                                    )
    yield from get_source_from_stream(pipe_iterator, state)  # type: ignore



from typing import Any, Iterator

from dlt.common.typing import DictStrAny
from dlt.pipeline.exceptions import MissingDependencyException

try:
    from oauth2client.service_account import ServiceAccountCredentials
    from apiclient.discovery import build
except ImportError:
    raise MissingDependencyException("Google Sheets Source", ["google1", "google2"])


# gets spreadsheet values from a single sheet preserving original typing for schema generation
# TODO: write schema discovery to better handle dates
# TODO: consider using https://github.com/burnash/gspread for spreadsheet discovery


def get_source(credentials: ServiceAccountCredentials,  spreadsheet_id: str, sheet_name: str) -> Iterator[DictStrAny]:
    sheets = _initialize_sheets(credentials)

    # get list of list of typed values
    result = sheets.spreadsheets().values().get(
        spreadsheetId=spreadsheet_id,
        range=sheet_name,
        # unformatted returns typed values
        valueRenderOption="UNFORMATTED_VALUE",
        # will return formatted dates
        dateTimeRenderOption="FORMATTED_STRING"
    ).execute()

    values = result.get('values')

    # yield dicts assuming row 0 contains headers and following rows values and all rows have identical length
    for v in values[1:]:
        yield {h: v for h, v in zip(values[0], v)}


def _initialize_sheets(credentials: ServiceAccountCredentials) -> Any:

    # Build the service object.
    service = build('sheets', 'v4', credentials=credentials)

    return service



from typing import Iterator, Sequence, Union, Any

from dlt.common.typing import DictStrAny
from dlt.pipeline.exceptions import MissingDependencyException


try:
    # import gracefully and produce nice exception that explains the user what to do
    import pandas
except ImportError:
    raise MissingDependencyException("SQL Query Source", ["pandas"], "SQL Query Source temporarily uses pandas as DB interface")

try:
    from sqlalchemy.exc import NoSuchModuleError
except ImportError:
    raise MissingDependencyException("SQL Query Source", ["sqlalchemy"], "SQL Query Source temporarily uses pandas as DB interface")


def get_source(
    sql: Any,
    con: Any,
    index_col: Union[str, Sequence[str], None] = None,
    coerce_float: bool = True,
    params: Any = None,
    parse_dates: Any = None,
    columns: Any = None,
    chunk_size: int = 1000
) -> Iterator[DictStrAny]:

    try:
        items = pandas.read_sql(sql, con, index_col=index_col, coerce_float=coerce_float, params=params, parse_dates=parse_dates, columns=columns, chunksize=chunk_size)
    except NoSuchModuleError:
        if "redshift.redshift_connector":
            raise MissingDependencyException("SQL Query Source", ["sqlalchemy-redshift", "redshift_connector"], "Redshift dialect support for SqlAlchemy")
        raise

    for i in items:
        record = i.to_dict("records")
        if len(record) > 0:
            yield record[0]



from typing import Any, Iterator, Union, cast, Sequence
from hexbytes import HexBytes
import requests

from dlt.common.typing import DictStrAny
from dlt.common.sources import TDeferred, TItem, defer_iterator, with_retry

from dlt.pipeline.exceptions import MissingDependencyException

try:
    # import gracefully and produce nice exception that explains the user what to do
    from web3 import Web3, HTTPProvider
except ImportError:
    raise MissingDependencyException("Ethereum Source", ["web3"], "Web3 is a all purpose python library to interact with Ethereum-compatible blockchains.")

# this source gets full state of any Ethereum compatible blockchain.
# - can be easily scaled up to hundreds of parallel extractions (ie. via Airflow)
# - supports retrying if node fails
# - supports multi threaded extraction on single machine via get_deferred_source
# - supports pipeline state so it may be used to just get new blocks when they are available
# - the return value is iterator (or deferred iterator) so still mappings may be used (ie. to decode transactions and logs), see example


def get_source(node_url: str, no_blocks: int, last_block: int = None, lag: int = 2, state: DictStrAny = None) -> Iterator[DictStrAny]:
    return _get_source(False, node_url, no_blocks, last_block, lag, state)  # type: ignore


def get_deferred_source(node_url: str, no_blocks: int, last_block: int = None, lag: int = 2, state: DictStrAny = None) -> Iterator[TDeferred[DictStrAny]]:
    return _get_source(True, node_url, no_blocks, last_block, lag, state)  # type: ignore


def _get_source(is_deferred: bool, node_url: str, no_blocks: int, last_block: int, lag: int, state: DictStrAny) -> Union[Iterator[TItem], Iterator[TDeferred[DictStrAny]]]:

    # this code is run only once
    w3 = Web3(Web3.HTTPProvider(node_url))

    # last block is not provided then take the highest block from the chain
    if last_block is None:
        # 14788082
        last_block = w3.eth.get_block_number() - lag

    # get default current block
    current_block = last_block - no_blocks + 1

    # get current block from the state if available
    if state:
        current_block = state.get("ethereum_current_block", current_block)

    assert current_block >= 0
    assert current_block <= last_block

    # get chain id
    chain_id = w3.eth.chain_id

    # do not get highest block (in fact we should have 3 block lag here)
    while current_block <= last_block:
        # get rid of AttributeDict (web3 didn't adopt TypedDict)
        print(f"requesting block {current_block}")

        @defer_iterator
        @with_retry()
        def _get_block_deferred() -> DictStrAny:
            return _get_block(w3, current_block, chain_id)

        @with_retry()
        def _get_block_retry() -> DictStrAny:
            return _get_block(w3, current_block, chain_id)

        # yield deferred items or actual item values
        if is_deferred:
            yield _get_block_deferred()
        else:
            yield _get_block_retry()
        current_block += 1

    # update state, it will be read and stored with the pipeline instance ONLY when whole iterator finishes
    # state participates in the same atomic operation as data
    # this must happen after last yield
    print(f"finalizing {state}")
    if state is not None:
        state["ethereum_current_block"] = current_block


def _get_block(w3: Web3, current_block: int, chain_id: int) -> DictStrAny:
    print(f"producing block {current_block}")

    # get block with all transaction
    block = dict(w3.eth.get_block(current_block, full_transactions=True))
    # set explicit chain id
    block["chain_id"] = chain_id
    # get rid of AttributeDict (web3 didn't adopt TypedDict)
    attr_txs = cast(Sequence[Any], block["transactions"])
    transactions: Sequence[DictStrAny] = [dict(tx) for tx in attr_txs]
    for tx in transactions:
        if "accessList" in tx and len(tx["accessList"]) > 0:
            tx["accessList"] = [dict(al) for al in tx["accessList"]]
        # propagate sorting and clustering info (we could also do it in unpacker: TODO:!)
        tx["block_timestamp"] = block["timestamp"]

        # overwrite chain_id which is not provided in all cases
        tx["chainId"] = chain_id

    block["transactions"] = transactions
    block["logsBloom"] = bytes(cast(HexBytes, block["logsBloom"]))  # serialize as bytes

    # get transaction receipts using batching. web3 does not support batching so we must
    # call node directly and then convert hex numbers to ints
    batch = []
    for idx, tx in enumerate(transactions):
        batch.append({
            "jsonrpc": "2.0",
            "method": "eth_getTransactionReceipt",
            "params": [tx["hash"]],
            "id": idx
        })
    provider = cast(HTTPProvider, w3.provider)
    r = requests.post(provider.endpoint_uri, json=batch, timeout=(20, 12))
    r.raise_for_status()
    receipts = r.json()
    for tx_receipt, tx in zip(receipts, transactions):
        if "result" not in tx_receipt:
            raise ValueError(tx_receipt)
        tx_receipt = tx_receipt["result"]
        assert tx_receipt["transactionHash"] == tx["hash"].hex()
        tx["transactionIndex"] = Web3.toInt(hexstr=tx_receipt["transactionIndex"])
        tx["status"] = Web3.toInt(hexstr=tx_receipt["status"])
        tx["logs"] = tx_receipt["logs"]
        for log in tx_receipt["logs"]:
            log["blockNumber"] = Web3.toInt(hexstr=log["blockNumber"])
            log["transactionIndex"] = Web3.toInt(hexstr=log["transactionIndex"])
            log["logIndex"] = Web3.toInt(hexstr=log["logIndex"])
            log["block_timestamp"] = block["timestamp"]
            log["block_hash"] = block["hash"]

    return block




from dataclasses import dataclass
from typing import Literal
from dlt.common import json

from dlt.common.typing import StrAny
from dlt.common.configuration.utils import TConfigSecret

TLoaderType = Literal["gcp", "redshift"]
TPipelineStage = Literal["extract", "unpack", "load"]

# extractor generator yields functions that returns list of items of the type (table) when called
# this allows generator to implement retry logic
# TExtractorItem = Callable[[], Iterator[StrAny]]
# # extractor generator yields tuples: (type of the item (table name), function defined above)
# TExtractorItemWithTable = Tuple[str, TExtractorItem]
# TExtractorGenerator = Callable[[DictStrAny], Iterator[TExtractorItemWithTable]]


@dataclass
class PipelineCredentials:
    CLIENT_TYPE: TLoaderType

    @property
    def schema_prefix(self) -> str:
        pass

    @schema_prefix.setter
    def schema_prefix(self, new_value: str) -> None:
        pass

@dataclass
class GCPPipelineCredentials(PipelineCredentials):
    PROJECT_ID: str
    DATASET: str
    BQ_CRED_CLIENT_EMAIL: str
    BQ_CRED_PRIVATE_KEY: TConfigSecret = None
    TIMEOUT: float = 30.0

    @property
    def schema_prefix(self) -> str:
        return self.DATASET

    @schema_prefix.setter
    def schema_prefix(self, new_value: str) -> None:
        self.DATASET = new_value

    @classmethod
    def from_services_dict(cls, services: StrAny, dataset_prefix: str) -> "GCPPipelineCredentials":
        assert dataset_prefix is not None

        return cls("gcp", services["project_id"], dataset_prefix, services["client_email"], services["private_key"])

    @classmethod
    def from_services_file(cls, services_path: str, dataset_prefix: str) -> "GCPPipelineCredentials":
        with open(services_path, "r", encoding="utf-8") as f:
            services = json.load(f)
        return GCPPipelineCredentials.from_services_dict(services, dataset_prefix)


@dataclass
class PostgresPipelineCredentials(PipelineCredentials):
    PG_DATABASE_NAME: str
    PG_SCHEMA_PREFIX: str
    PG_USER: str
    PG_HOST: str
    PG_PASSWORD: TConfigSecret = None
    PG_PORT: int = 5439
    PG_CONNECTION_TIMEOUT: int = 15

    @property
    def schema_prefix(self) -> str:
        return self.PG_SCHEMA_PREFIX

    @schema_prefix.setter
    def schema_prefix(self, new_value: str) -> None:
        self.PG_SCHEMA_PREFIX = new_value



from dlt.pipeline import Pipeline,  GCPPipelineCredentials

# credentials = PostgresPipelineCredentials("redshift", "chat_analytics_rasa", "mainnet_6", "loader", "3.73.90.3")
credentials = GCPPipelineCredentials.from_services_file("_secrets/project1234_service.json", "mainnet_4")
pipeline = Pipeline("ethereum")
pipeline.create_pipeline(credentials)
schema = Pipeline.load_schema_from_file("examples/schemas/ethereum_schema.yml")
# set the loaded schema for the whole pipeline
pipeline.set_default_schema(schema)
# will sync schema with the target
# pipeline.sync_schema()
# with pipeline.sql_client() as c:
#     c.execute_sql("BEGIN TRANSACTION;\n INSERT INTO `mainnet_6_ethereum._dlt_loads` VALUES('2022-06-09 22:09:21.254700 UTC', '1654812536.548904', 1); COMMIT TRANSACTION;")



from typing import Iterable, Union

from dlt.common.exceptions import DltException


class ConfigurationException(DltException):
    def __init__(self, msg: str) -> None:
        super().__init__(msg)


class ConfigEntryMissingException(ConfigurationException):
    """thrown when not all required config elements are present"""

    def __init__(self, missing_set: Iterable[str]) -> None:
        self.missing_set = missing_set
        super().__init__('Missing config keys: ' + str(missing_set))


class ConfigEnvValueCannotBeCoercedException(ConfigurationException):
    """thrown when value from ENV cannot be coerced to hinted type"""

    def __init__(self, attr_name: str, env_value: str, hint: type) -> None:
        self.attr_name = attr_name
        self.env_value = env_value
        self.hint = hint
        super().__init__('env value %s cannot be coerced into type %s in attr %s' % (env_value, str(hint), attr_name))


class ConfigIntegrityException(ConfigurationException):
    """thrown when value from ENV cannot be coerced to hinted type"""

    def __init__(self, attr_name: str, env_value: str, info: Union[type, str]) -> None:
        self.attr_name = attr_name
        self.env_value = env_value
        self.info = info
        super().__init__('integrity error for attr %s with value %s. %s.' % (attr_name, env_value, info))


class ConfigFileNotFoundException(ConfigurationException):
    """thrown when configuration file cannot be found in config folder"""

    def __init__(self, path: str) -> None:
        super().__init__(f"Missing config file in {path}")



from collections import abc
from functools import wraps
from typing import Any, Callable, Iterable, Iterator, Literal, Optional, Sequence, TypeVar, Union, TypedDict, cast
try:
    from typing_extensions import ParamSpec
except ImportError:
    ParamSpec = lambda x: [x]  # type: ignore

from dlt.common import logger
from dlt.common.time import sleep
from dlt.common.typing import DictStrAny, StrAny


# possible types of items yielded by the source
# 1. document (mapping from str to any type)
# 2. Iterable (ie list) on the mapping above for returning many documents with single yield
TItem = Union[DictStrAny, Sequence[DictStrAny]]
TBoundItem = TypeVar("TBoundItem", bound=TItem)
TDeferred = Callable[[], TBoundItem]

_TFunParams = ParamSpec("_TFunParams")

# name of dlt metadata as part of the item
DLT_METADATA_FIELD = "_dlt_meta"


class TEventDLTMeta(TypedDict, total=False):
    table_name: str  # a root table to which unpack the event


def append_dlt_meta(item: TBoundItem, name: str, value: Any) -> TBoundItem:
    if isinstance(item, abc.Sequence):
        for i in item:
            i.setdefault(DLT_METADATA_FIELD, {})[name] = value
    elif isinstance(item, dict):
        item.setdefault(DLT_METADATA_FIELD, {})[name] = value

    return item


def with_table_name(item: TBoundItem, table_name: str) -> TBoundItem:
    # normalize table name before adding
    return append_dlt_meta(item, "table_name", table_name)


def get_table_name(item: StrAny) -> Optional[str]:
    if DLT_METADATA_FIELD in item:
        meta: TEventDLTMeta = item[DLT_METADATA_FIELD]
        return meta.get("table_name", None)
    return None


def with_retry(max_retries: int = 3, retry_sleep: float = 1.0) -> Callable[[Callable[_TFunParams, TBoundItem]], Callable[_TFunParams, TBoundItem]]:

    def decorator(f: Callable[_TFunParams, TBoundItem]) -> Callable[_TFunParams, TBoundItem]:

        def _wrap(*args: Any, **kwargs: Any) -> TBoundItem:
            attempts = 0
            while True:
                try:
                    return f(*args, **kwargs)
                except Exception as exc:
                    if attempts == max_retries:
                        raise
                    attempts += 1
                    logger.warning(f"Exception {exc} in iterator, retrying {attempts} / {max_retries}")
                    sleep(retry_sleep)

        return _wrap

    return decorator


def defer_iterator(f: Callable[_TFunParams, TBoundItem]) -> Callable[_TFunParams, TDeferred[TBoundItem]]:

    @wraps(f)
    def _wrap(*args: Any, **kwargs: Any) -> TDeferred[TBoundItem]:
        def _curry() -> TBoundItem:
            return f(*args, **kwargs)
        return _curry

    return _wrap



from typing import Dict, Mapping, Optional, Sequence, Tuple, cast, TypedDict, Any

from dlt.common.schema import Schema
from dlt.common.schema.typing import TColumn, TColumnName, TSimpleRegex
from dlt.common.schema.utils import column_name_validator
from dlt.common.utils import uniq_id, digest128
from dlt.common.typing import DictStrAny, DictStrStr, TEvent, StrAny
from dlt.common.normalizers.json import TUnpackedRowIterator
from dlt.common.sources import DLT_METADATA_FIELD, TEventDLTMeta, get_table_name
from dlt.common.validation import validate_dict


class TEventRow(TypedDict, total=False):
    _dlt_id: str  # unique id of current row


class TEventRowRoot(TEventRow, total=False):
    _dlt_load_id: str  # load id to identify records loaded together that ie. need to be processed
    _dlt_meta: TEventDLTMeta  # stores metadata, should never be sent to the normalizer


class TEventRowChild(TEventRow, total=False):
    _dlt_root_id: str  # unique id of top level parent
    _dlt_parent_id: str  # unique id of parent row
    _dlt_list_idx: int  # position in the list of rows
    value: Any  # for lists of simple types


class JSONNormalizerConfigPropagation(TypedDict, total=True):
    root: Optional[Mapping[str, TColumnName]]
    tables: Optional[Mapping[str, Mapping[str, TColumnName]]]


class JSONNormalizerConfig(TypedDict, total=True):
    generate_dlt_id: Optional[bool]
    max_nesting: Optional[int]
    propagation: Optional[JSONNormalizerConfigPropagation]


# for those paths the complex nested objects should be left in place
def _is_complex_type(schema: Schema, table_name: str, field_name: str, _r_lvl: int) -> bool:
    # turn everything at the recursion level into complex type
    max_nesting  = schema._normalizers_config["json"].get("config", {}).get("max_nesting", 1000)
    assert _r_lvl <= max_nesting
    if _r_lvl == max_nesting:
        return True
    # or use definition in the schema
    column: TColumn = None
    table = schema._schema_tables.get(table_name)
    if table:
        column = table["columns"].get(field_name, None)
    if column is None:
        data_type = schema.get_preferred_type(field_name)
    else:
        data_type = column["data_type"]
    return data_type == "complex"


def _flatten(schema: Schema, table: str, dict_row: TEventRow, _r_lvl: int) -> Tuple[TEventRow, Dict[str, Sequence[Any]]]:
    out_rec_row: DictStrAny = {}
    out_rec_list: Dict[str, Sequence[Any]] = {}

    def norm_row_dicts(dict_row: StrAny, __r_lvl: int, parent_name: Optional[str]) -> None:
        for k, v in dict_row.items():
            corrected_k = schema.normalize_column_name(k)
            child_name = corrected_k if not parent_name else schema.normalize_make_path(parent_name, corrected_k)
            # for lists and dicts we must check if type is possibly complex
            if isinstance(v, dict) or isinstance(v, list):
                if not _is_complex_type(schema, table, child_name, __r_lvl):
                    if isinstance(v, dict):
                        # flatten the dict more
                        norm_row_dicts(v, __r_lvl + 1, parent_name=child_name)
                    else:
                        # pass the list to out_rec_list
                        out_rec_list[child_name] = v
                    continue
                else:
                    # pass the complex value to out_rec_row
                    pass

            out_rec_row[child_name] = v

    norm_row_dicts(dict_row, _r_lvl, None)
    return cast(TEventRow, out_rec_row), out_rec_list


def _get_child_row_hash(parent_row_id: str, child_table: str, list_idx: int) -> str:
    # create deterministic unique id of the child row taking into account that all lists are ordered
    # and all child tables must be lists
    return digest128(f"{parent_row_id}_{child_table}_{list_idx}")


def _add_linking(row: TEventRowChild, extend: DictStrAny, parent_row_id: str, list_idx: int) -> TEventRowChild:
        row["_dlt_parent_id"] = parent_row_id
        row["_dlt_list_idx"] = list_idx
        row.update(extend)  # type: ignore

        return row


def _get_content_hash(schema: Schema, table: str, row: StrAny) -> str:
    return digest128(uniq_id())


def _get_propagated_values(schema: Schema, table: str, row: TEventRow, is_top_level: bool) -> StrAny:
    config: JSONNormalizerConfigPropagation = schema._normalizers_config["json"].get("config", {}).get("propagation", None)
    extend: DictStrAny = {}
    if config:
        # mapping(k:v): propagate property with name "k" as property with name "v" in child table
        mappings: DictStrStr = {}
        if is_top_level:
            mappings.update(config.get("root", {}))
        if table in config.get("tables", {}):
            mappings.update(config["tables"][table])
        # look for keys and create propagation as values
        for prop_from, prop_as in mappings.items():
            if prop_from in row:
                extend[prop_as] = row[prop_from]  # type: ignore

    return extend


# generate child tables only for lists
def _normalize_list(
    schema: Schema,
    seq: Sequence[Any],
    extend: DictStrAny,
    table: str,
    parent_table: str,
    parent_row_id: Optional[str] = None,
    _r_lvl: int = 0
) -> TUnpackedRowIterator:

    v: TEventRowChild = None
    for idx, v in enumerate(seq):
        # yield child table row
        if isinstance(v, dict):
            yield from _normalize_row(schema, v, extend, table, parent_table, parent_row_id, idx, _r_lvl)
        elif isinstance(v, list):
            # unpack lists of lists, we assume all lists in the list have the same type so they should go to the same table
            list_table_name = schema.normalize_make_path(table, "list")
            yield from _normalize_list(schema, v, extend, list_table_name, parent_table, parent_row_id, _r_lvl + 1)
        else:
            # list of simple types
            child_row_hash = _get_child_row_hash(parent_row_id, table, idx)
            e = _add_linking({"value": v, "_dlt_id": child_row_hash}, extend, parent_row_id, idx)
            yield (table, parent_table), e


def _normalize_row(
    schema: Schema,
    dict_row: TEventRow,
    extend: DictStrAny,
    table: str,
    parent_table: Optional[str] = None,
    parent_row_id: Optional[str] = None,
    pos: Optional[int] = None,
    _r_lvl: int = 0
) -> TUnpackedRowIterator:

    is_top_level = parent_table is None
    # flatten current row and extract all lists to recur into
    flattened_row, lists = _flatten(schema, table, dict_row, _r_lvl)
    # infer record hash or leave existing primary key if present
    row_id = flattened_row.get("_dlt_id", None)
    if not row_id:
        # check if we have primary key: if so use it
        primary_key = schema.filter_row_with_hint(table, "primary_key", flattened_row)
        if primary_key:
            # create row id from primary key
            row_id = digest128("_".join(map(lambda v: str(v), primary_key.values())))
        elif not is_top_level:
            # child table row deterministic hash
            row_id = _get_child_row_hash(parent_row_id, table, pos)
            # link to parent table
            _add_linking(cast(TEventRowChild, flattened_row), extend, parent_row_id, pos)
        else:
            # create hash based on the content of the row
            row_id = _get_content_hash(schema, table, flattened_row)
        flattened_row["_dlt_id"] = row_id

    # find fields to propagate to child tables in config
    extend.update(_get_propagated_values(schema, table, flattened_row, is_top_level))

    # yield parent table first
    yield (table, parent_table), flattened_row

    # normalize and yield lists
    for k, list_content in lists.items():
        yield from _normalize_list(schema, list_content, extend, schema.normalize_make_path(table, k), table, row_id, _r_lvl + 1)


def extend_schema(schema: Schema) -> None:
    # validate config
    config = schema._normalizers_config["json"].get("config", {})
    validate_dict(JSONNormalizerConfig, config, "./normalizers/json/config", validator_f=column_name_validator(schema.normalize_column_name))

    # quick check to see if hints are applied
    default_hints = schema.schema_settings.get("default_hints", {})
    if "not_null" in default_hints and "^_dlt_id$" in default_hints["not_null"]:
        return
    # add hints
    schema.merge_hints(
        {
            "not_null": [
                TSimpleRegex("_dlt_id"), TSimpleRegex("_dlt_root_id"), TSimpleRegex("_dlt_parent_id"),
                TSimpleRegex("_dlt_list_idx"), TSimpleRegex("_dlt_load_id")
                ],
            "foreign_key": [TSimpleRegex("_dlt_parent_id")],
            "unique": [TSimpleRegex("_dlt_id")]
        }
    )


def normalize(schema: Schema, source_event: TEvent, load_id: str) -> TUnpackedRowIterator:
    # we will extend event with all the fields necessary to load it as root row
    event = cast(TEventRowRoot, source_event)
    # identify load id if loaded data must be processed after loading incrementally
    event["_dlt_load_id"] = load_id
    # find table name
    table_name = schema.normalize_table_name(get_table_name(event) or schema.schema_name)
    # drop dlt metadata before normalizing
    event.pop(DLT_METADATA_FIELD, None)  # type: ignore
    # use event type or schema name as table name, request _dlt_root_id propagation
    yield from _normalize_row(schema, cast(TEventRowChild, event), {}, table_name)



class LoadingVolumeConfiguration:
    LOADING_VOLUME_PATH: str = "_storage/loading"  # path to volume where files to be loaded to analytical storage are stored
    DELETE_COMPLETED_JOBS: bool = False  # if set to true the folder with completed jobs will be deleted

class ProductionLoadingVolumeConfiguration(LoadingVolumeConfiguration):
    LOADING_VOLUME_PATH: str = None



from typing import Iterator
from dlt.common.sources import with_table_name

from dlt.common.typing import DictStrAny
from dlt.common.time import timestamp_within


def get_source(base_source: Iterator[DictStrAny], source_env: str = None, start_timestamp: float = None, end_timestamp: float = None, state: DictStrAny = None) -> Iterator[DictStrAny]:
    # the base source can be a file, database table (see sql_query.py), kafka topic, rabbitmq queue etc.
    # recover start_timestamp from state if given
    if state:
        start_timestamp = max(start_timestamp or 0, state.get("rasa_tracker_store", 0))
    # we expect tracker store events here
    last_timestamp: int = None
    for source_event in base_source:
        # filter out events
        if timestamp_within(source_event["timestamp"], start_timestamp, end_timestamp):
            # yield tracker table with all-event index
            event_type = source_event["event"]
            last_timestamp = source_event["timestamp"]
            event = {
                "sender_id": source_event["sender_id"],
                "timestamp": last_timestamp,
                "event": event_type
            }
            if source_env:
                event["source"] = source_env
            if "model_id" in source_event:
                event["model_id"] = source_event["model_id"]
            yield with_table_name(event, "event")
            # yield original event
            yield with_table_name(source_event, "event_" + event_type)

    # write state so we can restart
    if state and last_timestamp:
        state["rasa_tracker_store"] = last_timestamp



from abc import ABC, abstractmethod
from types import TracebackType
from typing import Any, Generic, Iterator, List, Tuple, Optional, Sequence, Type, TypeVar, AnyStr, TypedDict
from pathlib import Path

from dlt.common import pendulum, logger
from dlt.common.dataset_writers import TWriterType
from dlt.common.schema import TColumn, Schema, TTableColumns

from dlt.loaders.local_types import LoadJobStatus
from dlt.loaders.exceptions import LoadClientSchemaVersionCorrupted, LoadUnknownTableException

# native connection
TNativeConn = TypeVar("TNativeConn", bound="object")

# type for dbapi cursor
class DBCursor:
    closed: Any
    connection: Any
    query: Any
    description: Tuple[Any, ...]

    def execute(self, query: AnyStr, *args: Any, **kwargs: Any ) -> None:
        ...
    def fetchall(self) -> List[Tuple[Any, ...]]:
        ...
    def fetchmany(self, size: int = ...) -> List[Tuple[Any, ...]]:
        ...
    def fetchone(self) -> Optional[Tuple[Any, ...]]:
        ...


class TJobClientCapabilities(TypedDict):
    writer_type: TWriterType


class LoadJob:
    def __init__(self, file_name: str) -> None:
        """
        File name is also a job id (or job id is deterministically derived) so it must be globally unique
        """
        self._file_name = file_name

    @abstractmethod
    def status(self) -> LoadJobStatus:
        pass

    @abstractmethod
    def file_name(self) -> str:
        pass

    @abstractmethod
    def exception(self) -> str:
        pass


class LoadEmptyJob(LoadJob):
    def __init__(self, file_name: str, status: LoadJobStatus, exception: str = None) -> None:
        self._status = status
        self._exception = exception
        super().__init__(file_name)

    def status(self) -> LoadJobStatus:
        return self._status

    def file_name(self) -> str:
        return self._file_name

    def exception(self) -> str:
        return self._exception


class JobClientBase(ABC):
    def __init__(self, schema: Schema) -> None:
        self.schema = schema

    @abstractmethod
    def initialize_storage(self) -> None:
        pass

    @abstractmethod
    def update_storage_schema(self) -> None:
        pass

    @abstractmethod
    def start_file_load(self, table_name: str, file_path: str) -> LoadJob:
        pass

    @abstractmethod
    def restore_file_load(self, file_path: str) -> LoadJob:
        pass

    @abstractmethod
    def complete_load(self, load_id: str) -> None:
        pass

    @property
    @abstractmethod
    def capabilities(self) -> TJobClientCapabilities:
        pass

    @abstractmethod
    def __enter__(self) -> "JobClientBase":
        pass

    @abstractmethod
    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType) -> None:
        pass

    def _get_table_by_name(self, table_name: str, file_name: str) -> TTableColumns:
        try:
            return self.schema.get_table_columns(table_name)
        except KeyError:
            raise LoadUnknownTableException(table_name, file_name)

    @staticmethod
    def get_file_name_from_file_path(file_path: str) -> str:
        return Path(file_path).name

    @staticmethod
    def make_job_with_status(file_path: str, status: LoadJobStatus, message: str = None) -> LoadJob:
        return LoadEmptyJob(JobClientBase.get_file_name_from_file_path(file_path), status, exception=message)

    @staticmethod
    def make_absolute_path(file_path: str) -> str:
        return str(Path(file_path).absolute())


class SqlClientBase(ABC, Generic[TNativeConn]):
    def __init__(self, default_schema_name: str) -> None:
        if not default_schema_name:
            raise ValueError(default_schema_name)
        self._qualified_default_schema_name: str = None
        self._qualified_default_schema_name = self.fully_qualified_schema_name(default_schema_name)

    @abstractmethod
    def open_connection(self) -> None:
        pass

    @abstractmethod
    def close_connection(self) -> None:
        pass

    def __enter__(self) -> "SqlClientBase[TNativeConn]":
        self.open_connection()
        return self

    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType) -> None:
        self.close_connection()

    @abstractmethod
    def native_connection(self) -> TNativeConn:
        pass

    @abstractmethod
    def has_schema(self, qualified_schema_name: str = None) -> bool:
        pass

    @abstractmethod
    def create_schema(self, qualified_schema_name: str = None) -> None:
        pass

    @abstractmethod
    def drop_schema(self, qualified_schema_name: str = None) -> None:
        pass

    @abstractmethod
    def execute_sql(self, sql: AnyStr, *args: Any, **kwargs: Any) -> Optional[Sequence[Sequence[Any]]]:
        pass

    @abstractmethod
    def execute_query(self, query: AnyStr, *args: Any, **kwargs: Any) -> Iterator[DBCursor]:
        pass

    @abstractmethod
    def fully_qualified_schema_name(self, schema_name: str = None) -> str:
        pass

    def fully_qualified_table_name(self, table_name: str) -> str:
        return f"{self.fully_qualified_schema_name()}.{table_name}"


class SqlJobClientBase(JobClientBase):
    def __init__(self, schema: Schema, sql_client: SqlClientBase[TNativeConn]) -> None:
        super().__init__(schema)
        self.sql_client = sql_client

    def update_storage_schema(self) -> None:
        storage_version = self._get_schema_version_from_storage()
        if storage_version < self.schema.schema_version:
            for sql in self._build_schema_update_sql():
                self.sql_client.execute_sql(sql)
            self._update_schema_version(self.schema.schema_version)

    def complete_load(self, load_id: str) -> None:
        name = self.sql_client.fully_qualified_table_name(Schema.LOADS_TABLE_NAME)
        now_ts = str(pendulum.now())
        self.sql_client.execute_sql(f"INSERT INTO {name}(load_id, status, inserted_at) VALUES('{load_id}', 0, '{now_ts}');")

    def __enter__(self) -> "SqlJobClientBase":
        self.sql_client.open_connection()
        return self

    def __exit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType) -> None:
        self.sql_client.close_connection()

    @abstractmethod
    def _build_schema_update_sql(self) -> List[str]:
        pass

    def _create_table_update(self, table_name: str, storage_table: TTableColumns) -> Sequence[TColumn]:
        # compare table with stored schema and produce delta
        updates = self.schema.get_schema_update_for(table_name, storage_table)
        logger.info(f"Found {len(updates)} updates for {table_name} in {self.schema.schema_name}")
        return updates

    def _get_schema_version_from_storage(self) -> int:
        name = self.sql_client.fully_qualified_table_name(Schema.VERSION_TABLE_NAME)
        rows = self.sql_client.execute_sql(f"SELECT {Schema.VERSION_COLUMN_NAME} FROM {name} ORDER BY inserted_at DESC LIMIT 1;")
        if len(rows) > 1:
            raise LoadClientSchemaVersionCorrupted(self.sql_client._qualified_default_schema_name)
        if len(rows) == 0:
            return 0
        return int(rows[0][0])

    def _update_schema_version(self, new_version: int) -> None:
        now_ts = str(pendulum.now())
        name = self.sql_client.fully_qualified_table_name(Schema.VERSION_TABLE_NAME)
        self.sql_client.execute_sql(f"INSERT INTO {name}({Schema.VERSION_COLUMN_NAME}, engine_version, inserted_at) VALUES ({new_version}, {Schema.ENGINE_VERSION}, '{now_ts}');")




from typing import Iterator
from dlt.pipeline import Pipeline
from dlt.pipeline.typing import GCPPipelineCredentials

from examples.sources.singer_tap import SingerMessage, get_source_from_stream
from examples.sources.jsonl import get_source as read_jsonl

# load hubspot schema stub - it converts all field names with `timestamp` into timestamp type
schema = Pipeline.load_schema_from_file("examples/schemas/hubspot_schema.yml")
p = Pipeline("singer_tap_hubspot")
p.create_pipeline(GCPPipelineCredentials.from_services_file("_secrets/project1234_service.json", "load_3"), schema=schema)
# p.extract(get_source(read_jsonl("examples/data/singer_taps/tap_google_sheet.jsonl")))
raw_messages: Iterator[SingerMessage] = read_jsonl("examples/data/singer_taps/tap_hubspot.jsonl")  # type: ignore
p.extract(get_source_from_stream(raw_messages, p.state))
p.unpack()
# print(p.get_default_schema().as_yaml(remove_defaults=True))
p.load()



import yaml
import os
import argparse
from typing import Callable

from dlt.common import json
from dlt.common.schema import Schema
from dlt.common.runners.pool_runner import TRunArgs, add_pool_cli_arguments
from dlt.common.typing import DictStrAny
from dlt.common.utils import str2bool


def str2bool_a(v: str) -> bool:
    try:
        return str2bool(v)
    except ValueError:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def main() -> None:
    parser = argparse.ArgumentParser(description="Runs various DLT modules", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(dest="command")
    unpack = subparsers.add_parser("unpack", help="Runs unpacker")
    add_pool_cli_arguments(unpack)
    load = subparsers.add_parser("load", help="Runs loader")
    add_pool_cli_arguments(load)
    dbt = subparsers.add_parser("dbt", help="Executes dbt package")
    add_pool_cli_arguments(dbt)
    schema = subparsers.add_parser("schema", help="Shows, converts and upgrades schemas")
    schema.add_argument("file", help="Schema file name, in yaml or json format, will autodetect based on extension")
    schema.add_argument("--format", choices=["json", "yaml"], default="yaml", help="Display schema in this format")
    schema.add_argument("--remove-defaults", action="store_true", help="Does not show default hint values")

    args = parser.parse_args()
    run_f: Callable[[TRunArgs], None] = None

    if args.command == "unpack":
        from dlt.unpacker.unpacker import run_main as unpacker_run
        run_f = unpacker_run
    elif args.command == "load":
        from dlt.loaders.loader import run_main as loader_run
        run_f = loader_run
    elif args.command == "dbt":
        from dlt.dbt_runner.runner import run_main as dbt_run
        run_f = dbt_run
    elif args.command == "schema":
        with open(args.file, "r", encoding="utf-8") as f:
            if os.path.splitext(args.file)[1][1:] == "json":
                schema_dict: DictStrAny = json.load(f)
            else:
                schema_dict = yaml.safe_load(f)
        s = Schema.from_dict(schema_dict)
        if args.format == "json":
            schema_str = json.dumps(s.to_dict(remove_defaults=args.remove_defaults), indent=2)
        else:
            schema_str = s.as_yaml(remove_defaults=args.remove_defaults)
        print(schema_str)
        exit(0)
    else:
        parser.print_help()
        exit(-1)

    run_args = TRunArgs(args.single_run, args.wait_runs)
    run_f(run_args)

if __name__ == "__main__":
    main()



from typing import List, Sequence, Tuple, Type
from itertools import groupby
from pathlib import Path

from dlt.common.utils import chunks
from dlt.common.file_storage import FileStorage
from dlt.common.configuration import UnpackingVolumeConfiguration
from dlt.common.storages.versioned_storage import VersionedStorage


class UnpackerStorage(VersionedStorage):

    STORAGE_VERSION = "1.0.0"
    UNPACKING_FOLDER: str = "unpacking"  # folder within the volume where files to be unpacked are stored
    UNPACK_FILE_EXTENSION = ".unpack.json"
    UNPACK_FILE_EXTENSION_LEN = len(UNPACK_FILE_EXTENSION)

    def __init__(self, is_owner: bool, C: Type[UnpackingVolumeConfiguration]) -> None:
        super().__init__(UnpackerStorage.STORAGE_VERSION, is_owner, FileStorage(C.UNPACKING_VOLUME_PATH, "t", makedirs=is_owner))

    def initialize_storage(self) -> None:
        self.storage.create_folder(UnpackerStorage.UNPACKING_FOLDER, exists_ok=True)

    def list_files_to_unpack_sorted(self) -> Sequence[str]:
        return sorted(self.storage.list_folder_files(UnpackerStorage.UNPACKING_FOLDER))

    def get_grouped_iterator(self, files: Sequence[str]) -> "groupby[str, str]":
        return groupby(files, lambda f: UnpackerStorage.get_schema_name(f))

    @staticmethod
    def chunk_by_events(files: Sequence[str], max_events: int, processing_cores: int) -> List[Sequence[str]]:
        # should distribute ~ N events evenly among m cores with fallback for small amounts of events

        def count_events(file_name : str) -> int:
            # return event count from file name
            return UnpackerStorage.get_events_count(file_name)

        counts = list(map(count_events, files))
        # make a list of files containing ~max_events
        events_count = 0
        m = 0
        while events_count < max_events and m < len(files):
            events_count += counts[m]
            m += 1
        processing_chunks = round(m / processing_cores)
        if processing_chunks == 0:
            # return one small chunk
            return [files]
        else:
            # should return ~ amount of chunks to fill all the cores
            return list(chunks(files[:m], processing_chunks))

    @staticmethod
    def get_events_count(file_name: str) -> int:
        return UnpackerStorage._parse_unpack_file_name(file_name)[0]

    @staticmethod
    def get_schema_name(file_name: str) -> str:
        return UnpackerStorage._parse_unpack_file_name(file_name)[2]

    @staticmethod
    def build_unpack_file_name(schema_name: str, stem: str, event_count: int, load_id: str) -> str:
        # builds file name of the unpack file for the tracker
        return f"{schema_name}_{stem}_{load_id}_{event_count}{UnpackerStorage.UNPACK_FILE_EXTENSION}"

    @staticmethod
    def _parse_unpack_file_name(file_name: str) -> Tuple[int, str, str]:
        # parser unpack tracker file and returns (events found, load id, schema_name)
        if not file_name.endswith(UnpackerStorage.UNPACK_FILE_EXTENSION):
            raise ValueError(file_name)

        parts = Path(file_name[:-UnpackerStorage.UNPACK_FILE_EXTENSION_LEN]).stem.split("_")
        return (int(parts[-1]), parts[-2], parts[0])



from dlt.common import json
from dlt.common.schema import Schema
from dlt.common.typing import DictStrAny

from dlt.pipeline import Pipeline, PostgresPipelineCredentials

# this is example of extracting from iterator that may be a
# - jsonl file you read line by line
# - a list of dicts
# - iterator over panda frame https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_dict.html
# - any iterator with map function
# - any iterator whatsoever ie getting data from REST API

# extracting from iterator always happens in single worker thread and iterator is directly returning data so there's no retry mechanism if anything fails

# working BQ creds
# credentials = Pipeline.load_gcp_credentials("_secrets/project1234_service.json", "gamma_guild")

if __name__ == '__main__':
    # working redshift creds, you can pass password as last parameter or via PG_PASSWORD env variable ie.
    # LOG_LEVEL=INFO PG_PASSWORD=.... python examples/discord_iterator.py
    credentials = PostgresPipelineCredentials("redshift", "chat_analytics_rasa", "gamma_guild_8", "loader", "3.73.90.3")

    pipeline = Pipeline("discord")
    schema: Schema = None
    # uncomment to use already modified schema, but the auto-inferred schema will also work nicely
    # TODO: schema below needs is not yet finished, more exploration needed
    # schema = Pipeline.load_schema_from_file("examples/schemas/discord_schema.yml")
    pipeline.create_pipeline(credentials, schema=schema)
    # the pipeline created a working directory. you can always attach the pipeline back by just providing the dir to
    # Pipeline::restore_pipeline
    print(pipeline.root_path)

    # load our sample data - list of discord channels
    with open("examples/data/channels.json", "r", encoding="utf-8") as f:
        channels = json.load(f)
    # and extract it
    pipeline.extract(channels, table_name="channels")
    # load list of messages
    with open("examples/data/messages.json", "r", encoding="utf-8") as f:
        messages = json.load(f)

    # but we actually want to process messages before processing so pass mapping function instead
    def processor(m: DictStrAny) -> DictStrAny:
        if "referenced_message" in m and m["referenced_message"] is not None:
            # we do not want full entity for referenced message, just the id
            m["referenced_message"] = m["referenced_message"]["id"]

        return m

    # pass mapping iterator over messages and extract
    pipeline.extract(map(processor, messages), table_name="messages")

    # from now on each pipeline does more or less the same thing: unpack and load data

    # now create loading packages and infer the schema
    pipeline.unpack()

    # show loads, each load contains a copy of the schema that corresponds to the data inside
    # and a set of directories for job states (new -> in progress -> failed|completed)
    new_loads = pipeline.list_unpacked_loads()

    # get inferred schema
    # schema = pipeline.get_current_schema()
    # print(schema.as_yaml(remove_default=True))

    # load packages
    pipeline.load()

    # should be empty
    new_loads = pipeline.list_unpacked_loads()
    print(new_loads)

    # now enumerate all complete loads if we have any failed packages
    # complete but failed job will not raise any exceptions
    completed_loads = pipeline.list_completed_loads()
    # print(completed_loads)
    for load_id in completed_loads:
        print(f"Checking failed jobs in {load_id}")
        for job, failed_message in pipeline.list_failed_jobs(load_id):
            print(f"JOB: {job}\nMSG: {failed_message}")



from dlt.common import Decimal
from dlt.common.time import sleep
from dlt.common.typing import DictStrAny
from dlt.pipeline import Pipeline, GCPPipelineCredentials
from dlt.common.arithmetics import numeric_default_context, numeric_default_quantize

from examples.schemas.ethereum import discover_schema
from examples.sources.ethereum import get_source

credentials = GCPPipelineCredentials.from_services_file("_secrets/project1234_service.json", "mainnet_6")
# credentials = PostgresPipelineCredentials("redshift", "chat_analytics_rasa", "mainnet_6", "loader", "3.73.90.3")

pipeline = Pipeline("ethereum")
pipeline.create_pipeline(credentials, schema=discover_schema())
print(pipeline.root_path)

# get ethereum source which is python iterator, request 3 newest blocks with default lag. also pass the state (optional)
# so you can get newly produced blocks when the pipeline is run again
i = get_source("https://mainnet.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161", 2, state=pipeline.state)


# you can transform the blocks by defining mapping function
def from_wei_to_eth(block: DictStrAny) -> DictStrAny:
    # convert wei to eth providing with numeric precision at the target
    with numeric_default_context():
        for tx in block["transactions"]:
            v_wei = tx["value"]
            if v_wei > 0:
                # 9 scale vs 18 digits ether scale: a no go for any financial operations
                # TODO: better support for decimal types
                tx["eth_value"] = numeric_default_quantize(Decimal(v_wei) / 10**18)
                # print(f"Produced nice decimal {tx['eth_value']}")
    return block


# extract data from the source. operation is atomic and like all atomic operations in pipeline, it does not raise but returns
# execution status
pipeline.extract(map(from_wei_to_eth, i), table_name="blocks")
print(pipeline.state)

# wait for ethereum network to produce some more blocks
sleep(20)

# restore the pipeline from the working directory (simulate continuation from the saved state)
pipeline.restore_pipeline(credentials, pipeline.root_path)
# obtain new iterator (the old one is expired), this time use deferred iterator to allow parallel block reading
i = get_source("https://mainnet.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161", 2, state=pipeline.state)
pipeline.extract(map(from_wei_to_eth, i), table_name="blocks")
print(pipeline.state)

# this will unpack and load all extracted data
pipeline.flush()
